/*
 *	rex.c - Rename EXtension of file
 *	  by Masaki NISHIMOTO
 *	  Usage: rex [-fivpeoa] .[extension] files...
 *	  ver.1.10
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>

#include "tools.h"

#define  BAD_FILE_NAME			"\"*/:<>?\\|"
#define  BAD_FILE_NAME_SFN		".,=[];"
#define  YES_OR_NO_P			yn_p_with_ret
/* #define	YES_OR_NO_P 		yes_or_no_p */
#define  TRUE					1
#define  FALSE					0
typedef int boolean;

char	*cmdname = "rex";

boolean	opt_errabort = FALSE;

int		error_code = 0;

#define  FATALEXIT 2

void expand_hook()
{
	_expand_abendcode = FATALEXIT;
	_dosfind_param = 0x25;
	expand_hook_default();
}

void help()
{
	fputss( stderr,
		"Usage: ", cmdname, " [-fivpeoa] .[extension] files...\n",
		"-f : rename readonly files without prompt\n",
		"-i : prompt whether to rename each file\n",
		"-v : print the name of each file\n",
		"-p : output the changed name of each file to stdout\n"
		"-e : abort if an error occurred\n",
		"-o : rename only non-extension files\n",
		"-a : append mode (LFN only)\n",
	NULL );
	exit( FATALEXIT );
}

void errors( char* filename )
{
	fputss( stderr, filename, ": ", sys_errlist[errno], ".\n", NULL );
	if ( opt_errabort )
		exit( error_code );
}

void rex_errors( char* filename, char* msg )
{
	fputss( stderr, filename, ": ", msg, ".\n", NULL );
	if ( opt_errabort )
		exit( error_code );
}

char* cut_last_periods( char* string )
{
	char* tmp = string + strlen( string ) - 1 ;
	if ( strlen( string ) <= 1 )
		return string;
	if ( *tmp == '.' )
	{
		(void)strcpy( tmp, "" );
		cut_last_periods( string );
	}
	else
		return string;
}

boolean isdotfile( char* name_, char* colon_, char* slash_, char* period_ )
{
	if ( ! slash_ )
	{
		if ( *name_ == '.' )
		{
			char* tmp = name_;
			while ( *tmp == '.' )
				tmp++;
			if ( period_ < tmp )
				return TRUE;
		}
		else
		{
			if ( colon_ )
			{
				if ( *( colon_ + 1) == '.' )
				{
					char* tmp = colon_ + 1;
					while ( *tmp == '.' )
						tmp++;
					if ( period_ < tmp )
						return TRUE;
				}
			}
		}
	}
	else
	{
		if ( *( slash_ + 1 ) == '.' )
		{
			char* tmp = slash_ + 1;
			while ( *tmp == '.' )
				tmp++;
			if ( period_ < tmp )
				return TRUE;
		}
	}
	return FALSE;
}

int main( int ac, char** av )
{
	char*		ext = NULL;
	char*		pt = NULL;
	char*		period = NULL;
	char*		slash = NULL;
	char*		colon = NULL;
	char*		oldname = NULL;
	char*		newname = NULL;
	int 		opt = 0;
	int			newextlen = 0;
	boolean 	opt_verbose = FALSE;
	boolean 	opt_force = FALSE;
	boolean 	opt_interactive = FALSE;
	boolean 	opt_noextonly = FALSE;
	boolean 	opt_changedname = FALSE;
	boolean		opt_append = FALSE;
	boolean		lfn = TRUE;
	unsigned	attr = 0;

	extern int	getopt();
	extern int	optind;
	extern int	opterr;
	extern char	*optarg;

#ifdef EXPLICIT_EXPARG
	_expand_abendcode = FATALEXIT;
	_dosfind_param = 0x25;
	ac = expand_arg( av[0], get_psp_arg(), &av );
#endif

	if ( ( _lfnsupp_level == LFN_NOSUPP ) || ( _lfnsupp_level == LFN_NONE ) )
		lfn = FALSE;
	else
		lfn = TRUE;

	opterr = 0;
	while ( ( opt = getopt( ac, av, "fivpeoa" ) ) != EOF )
	{
		switch ( opt )
		{
		case 'f':
			opt_force = TRUE;
			break;
		case 'i':
			opt_interactive = TRUE;
			break;
		case 'v':
			opt_verbose = TRUE;
			break;
		case 'p':
			opt_changedname = TRUE;
			break;
		case 'e':
			opt_errabort = TRUE;
			break;
		case 'o':
			opt_noextonly = TRUE;
			break;
		case 'a':
			if ( lfn ) 
				opt_append = TRUE;
			break;
		default:
			help();
		}
	}
	ac -= optind;
	av += optind;

	if ( **av == '.' )
	{
		ac--;
		newextlen = strlen( *av );
		ext = (char*)malloc( newextlen + 1 );
		(void)strncpy( ext, *av, newextlen );
		(void)strcpy( ext + newextlen, "" );
		ext = cut_last_periods( ext );
		if ( ( ! lfn ) && ( newextlen > 4 ) )
		{
			error_code = FATALEXIT;
			rex_errors( ext, "Extension too long for SFN" );
			exit( FATALEXIT );
		}
		pt = ext;
		while ( *++pt )
		{
			if ( strchr( BAD_FILE_NAME, *pt ) )
			{
				error_code = FATALEXIT;
				rex_errors( ext, "Bad extension" );
				exit( FATALEXIT );
			}
		}
		pt = ext;
		if ( ! lfn )
		{
			while ( *++pt )
			{
				if ( strchr( BAD_FILE_NAME_SFN, *pt ) )
				{
					error_code = FATALEXIT;
					rex_errors( ext, "Bad extension for SFN" );
					exit( FATALEXIT );
				}
			}
		}
	}
	else
		help();

	if ( ac == 0 )
		help();

	while ( ac-- > 0 )
	{
		char tmp;

		pt = *++av;
		regularize_pathname( pt );
		conv_to_unix_format( pt );
		pt = cut_last_periods( pt );
		oldname = strdup( pt );
		newname = (char*)malloc( strlen( pt ) + newextlen + 1 );
		tmp = *( oldname + strlen( oldname ) -1 );
		if ( ( tmp == '.' ) || ( tmp == '/' ) || ( tmp == ':' ) )
			goto cont;
		if ( ! _dos_getfileattr( oldname, &attr ) )
		{
			colon = strrchr( pt, ':' );
			slash = strrchr( pt, '/' );
			period = strrchr( pt, '.' );
			if ( ! isdotfile( pt, colon, slash, period ) )
			{
				if ( period && ( ! opt_append ) )
				{
					if ( ( slash && ( slash < period ) ) || ( ! slash ) )
					{
						if ( opt_noextonly )
							goto cont;
						(void)strcpy( period, "" );
					}
				}
			}
			(void)strcpy( newname, pt );
			(void)strcat( newname, ext );
			newname = cut_last_periods( newname );
			if ( ! access( newname, 0 ) )
			{
				error_code = 1;
				rex_errors( oldname, "Duplicate file name" );
				goto cont;
			}
			if ( opt_interactive )
			{
				fputss( stderr, "rename ", oldname, " ? ", NULL );
				if ( ! YES_OR_NO_P() )
					goto cont;
			}
			else
			{
				if ( ! ( opt_force ) && ( attr&_A_RDONLY ) )
				{
					fputss( stderr, oldname, " readonly, rename? ", NULL );
					if ( ! YES_OR_NO_P() )
						goto cont;
				}
			}
			if ( ! rename( oldname, newname ) )
			{
				if ( opt_verbose )
					fputss( stderr, oldname, " -> ", newname, "\n", NULL );
				if ( opt_changedname )
					fputss( stdout, newname, "\n", NULL );
			}
			else
			{
				error_code = 1;
				errors( oldname );
			}
		}
		else
		{
			error_code = 1;
			errors( oldname );
		}
cont:
	free( newname );
	free( oldname );
	}
	exit( error_code );
	/* G[̏ꍇA-e w肵Ȃ~͂ȂAIR[h͂1ԂB*/
}
