/**

	'xrm': another file delete utility for MS-DOS machines.

	rm command, source of xrm

	89.03.02	ver 1.0		Yasushi Saito, TAIST.
	90.10.03			Some cleanups,
					supports '.../' notation.
					-i asks y/n for all files.

        xrm.c has been modified from rm.c

	94.2.18         ver 0.1         test version
        95.8.18         ver 1.0         first release
        96.8.14                         this version

	                   compiled by LSI-C ver 3.30c trial version

	xrm [options] directory1 directory2 .. wildcardsn

	Remove except of file1, file2, ..., filen in directories

	Options:
	-r: Remove all files under directory, if matched file is a directory.
	-i: Interactive, i.e., ask yes or no when removing files.
	-f: Remove readonly files without query.
	-v: Verbose message.
	-t: Output filenames to stdout.
	-a: Hidden files and directories too.

	Wildcards can be used to give file or directory name ambiguously.
	See 'wild.man' for the wildcard specs.

	This program is NOT COPYRIGHTED, used, modified, re-distributed
		 with no restriction, and no warranty.

*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <dos.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>

#include "tools.h"

#define YES_OR_NO_P yn_p_with_ret	/* y/nœƂ^[v */

/* initially all 0 */
int             opt_recursive;
int             opt_verbose;
int             opt_interactive;
int             opt_force;
int             opt_test;
int             opt_all;

extern int      _dosfind_param;

int             _exclusive_exist;
unsigned        _file_attr;

/* WJÕfBNgƃt@Cw̍ő吔B */
#define MAX 64			/* 64 Ώ\낤B */
#define True 1
#define False 0

char           *MSG_ISDIR = "Is a directory";

#define FATALEXIT 2
#ifdef LSI_C
void
expand_control()
{
    _expand_abendcode = FATALEXIT;
}
#endif

static char    *msgs[] =
{
    "xrm [-rfivta] [pathes] wildcards\n",
    "-r : Recursively remove all files under directory\n",
    "-f : Remove readonly files without query\n",
    "     and ignore all errors\n",
    "-i : Query before removing\n",
    "-v : Verbose message\n",
    "-t : Output filenames into stdout\n",
    "-a : Hidden files and directories too\n",
};

int             error_code;

void
fputx(int i, FILE * f)
{
    if (i >> 4)
	fputx(i >> 4, f);
    i &= 0xf;
    putc(i + (i <= 10 ? '0' : 'a' - 10), f);
}

void
rm_error(char *s, char *reason)
{
    if (opt_force)
	return;
    fputss(stderr, s_conv_to_unix_format(s), ": ", reason, ".\n", NULL);
    error_code = 1;
}

void
io_error(char *msg, char *s)
{
    fputss(stderr, msg, s_conv_to_unix_format(s), ": ",
	   sys_errlist[errno], ".\n", NULL);
    if (!opt_force)
	return;
    error_code = 1;
}

/*
 * Actually remove PATH. PATH is a normal file (not directory).
 */

void
remove_file(char *path)
{
    if (opt_interactive) {
	fputss(stderr, "remove ", s_conv_to_unix_format(path), " ? ", NULL);
	if (!YES_OR_NO_P())
	    return;
    }
    if (remove(path) < 0) {
	unsigned        attr;
	if (_dos_getfileattr(path, &attr)) {
	    rm_error(path, sys_errlist[ENOENT]);
	    return;
	}
	if (attr & 0x10) {
	    rm_error(path, MSG_ISDIR);
	    return;
	}
	if (!opt_force) {
	    fputs("override protection 0x", stderr);
	    fputx(attr, stderr);
	    fputss(stderr, " for ", s_conv_to_unix_format(path), " ? ", NULL);
	    if (!YES_OR_NO_P())
		return;
	}
	_dos_setfileattr(path, attr & ~7);
	if (remove(path) < 0) {
	    io_error("Can't remove ", path);
	    _dos_setfileattr(path, attr);	/* nen no tame. Nide */
	    return;
	}
    }
    if (opt_verbose) {
	fputss(stderr, s_conv_to_unix_format(path), "\n", NULL);
    }
}

/* used as a SIGINT handler in delete_all_using_FCB() ONLY */
char           *restore_cwd;
void
inthndl()
{
#if 0
    fputs("Inturrupt\n", stderr);
#endif
    chdir(restore_cwd), exit(1);
}

/*
 * Remove all files under the directory S
 * 
 * Same as performing "DEL S\*.*"(S is the parameter) from commandline. Returns
 * immediately with files left untouched if failed to chdir S (hence S need
 * not a directory, in which case this procedure performs nothing).
 * 
 * Hidden files and readonly files may not be removed by this procedure, so this
 * is used as preprocess for whole-directory remove.
 * 
 */
delete_all_using_FCB(char *s)
{
    unsigned char   fcbbuf[37];
    unsigned char   cwdbuf[64];	/* MS-DOS̎dl64oCgł悢 */
    union REGS      in, out;
    void            (*orighndl) ();

    /* Set the drive name */
    if (isalpha(*s) && s[1] == ':') {
	if (s[0] >= 'a')
	    s[0] -= ('a' - 'A');
	fcbbuf[0] = s[0] - 'A' + 1;
	cwdbuf[0] = s[0];
	cwdbuf[1] = ':';
	cwdbuf[2] = '\\';
	in.x.si = (unsigned int) (cwdbuf + 3);
    } else {
	/* Current drive : fcbbuf[0]==0 */
	fcbbuf[0] = 0;
	cwdbuf[0] = '\\';
	in.x.si = (unsigned int) (cwdbuf + 1);
    }

    in.h.dl = fcbbuf[0];
    /*
     * Save the current directory. CWDBUF holds the current directory (of the
     * drive where S resides) after intdos call.
     */
    in.h.ah = 0x47;
    intdos(&in, &out);

    /*
     * S may not be a directory. Return immediately(with no errror) if failed
     * to chdir S.
     */
    restore_cwd = cwdbuf;
    if (SIG_IGN != (orighndl = signal(SIGINT, inthndl)))
	signal(SIGINT, inthndl);
    if (!chdir(s)) {
	/* delete all the files using BDOS call. */
	strcpy(&fcbbuf[1], "*       *  ");
	in.h.ah = 0x13;
	in.x.dx = (unsigned int) fcbbuf;
	intdos(&in, &out);

	/* Restore the current directory. */
	chdir(cwdbuf);
    }
    signal(SIGINT, orighndl);
}

/*
 * Delete the file PATH. Delete recursively if PATH is a directory and
 * OPT_RECURSIVE option is True.
 */
void
r_remove_file(char *path, unsigned attr)
{
    if (!(attr & 0x10)) {	/* fBNg֐exclusive
				 * ̂ -r w莞̂ */
	remove_file(path);
	return;
    } else
	/* PATH is directory. delete it and files under it. */
    {
	struct find_t   f;
	char            temp[PATHNAMELENGTH];
	/* First, perform BDOS file remove for speedups. */
	if (!opt_interactive) {
	    if (opt_verbose)
		fputss(stderr, s_conv_to_unix_format(path), "/*\n", NULL);
	    delete_all_using_FCB(path);
	}
	/*
	 * Delete remaining files(hidden files, readonly files) by searching
	 * each files.
	 */
	expand_path(temp, path, "*.*");
	if (_dos_findfirst(temp, 0x37, &f)) {
	    /*
	     * unable to interpret PATH as a directory. treat it as a plain
	     * file, and remove it
	     */
	    remove_file(path);
	    return;
	}
	do {
	    if (*(f.name) != '.') {
		expand_path(temp, path, f.name);
		if (f.attrib & 0x10)
		    r_remove_file(temp, f.attrib);
		else
		    remove_file(temp);
	    }
	} while (!_dos_findnext(&f));

	if (opt_interactive) {
	    fputss(stderr, "remove ",
		   s_conv_to_unix_format(path), "/ ? ", NULL);
	    if (!YES_OR_NO_P())
		return;
	}
	if (rmdir(path) < 0)
	    io_error("Can't remove directory ", path);
	if (opt_verbose)
	    fputss(stderr, s_conv_to_unix_format(path), "/\n", NULL);
    }
}

char           *
to_unix_format(char *path)
{
    unsigned        atr;

    _dos_getfileattr(path, &atr);
    conv_to_unix_format(path);

    if (atr & 0x10)
	strcat(path, "/\0");
    /*
     * \ۂɃfBNgƃt@C̋ʂt悤ɂB
     * ̊֐̈͏\ȑ傫z񂾂AyhłB
     */
    return path;
}

void
exclusive(char *dirs, WILD_INTERNAL wilds[MAX])
{
    int             i, j;

    char            buf[PATHNAMELENGTH];
    char            pathname[PATHNAMELENGTH];

    struct find_t   find;

    /* expand_wild œWJȂ΃fBNg̓WJ */
    if (!absolute_pathname(dirs, buf)) {
	expand_path(pathname, buf, "*.*");

	if (!_dos_xfindfirst(pathname, _file_attr, &find))
	    do {
		/* compare with all of wildcards */
		if (*find.name != '.')
		    for (i = j = 0; NULL != wilds[i].wildmark &&
			 !(j = name_match(find.name, wilds[i])); i++);

		if (!j && (*find.name != '_' || opt_all)) {
		    char            rm_file[PATHNAMELENGTH];
		    _exclusive_exist = True;

		    expand_path(rm_file, buf, find.name);
		    if (opt_test)
			fputss(stdout, to_unix_format(rm_file), "\n", NULL);
		    else
			r_remove_file(rm_file, find.attrib);
		}
	    } while (!_dos_findnext(&find));
    } else
	/* absolute_pathname œWJłȂ  pX */
	rm_error(conv_to_unix_format(dirs), "can't get the path");
}

void
usage(void)
{
    int             i;

    for (i = 0; i < sizeof(msgs) / sizeof(*msgs); i++)
	fputs(msgs[i], stderr);
    /* Nide, ANSIRpC̕ւ̂߂ɕAڂ߂ */
    exit(FATALEXIT);
}

int
r_exclusive(char **dirs, WILD_INTERNAL wilds[MAX], int t)
{
    opt_test = t;

    while (*dirs)
	exclusive(*dirs++, wilds);
    return _exclusive_exist;
}

int
check_path_or_file(char *argv)
{
    /*
     * ꕶڂAt@xbgłQڂ':'ȂfBNgB
     * MS-DOSł͂Q':'̃t@C͍ȂB
     * ܂Aꕶڂ'.'̃t@C̓fBNĝ݁B
     */

    if ((isalpha(*argv) && *(argv + 1) == ':') || *argv == '.')
	return True;

    while (*argv)
	if (*argv == PATH_SEP || *argv == '/')
	    return True;
	else
	    argv++;
    return False;
}

int
check_end_delimiter(char *path)
{
    while (*(++path));
    return (*(path - 1) == PATH_SEP || *(path - 1) == '/');
}

char           *
malloc_for_string(char *str)
{
    char           *pointer;
    void           *malloc(unsigned);

    if (pointer = (char *) malloc(sizeof(unsigned char) * strlen(str) +1))
	return pointer;

    _putnomemmes();
    exit(1);
}

int
main(int argc, char *argv[])
{
    int             i, c;
    int             pathes, wilds;

    char          **expanded_dirs;

    WILD_INTERNAL   wilds_entry[MAX];
    static char    *path_entry[MAX] =
    {
	"", "./", ""		/* current directory is default */
    };

    char            tmp[PATHNAMELENGTH];

    int             getopt(int, char **, char *);
    extern int      optind, opterr;
    extern char    *optarg;

    opterr = 0;

    while ((c = getopt(argc, argv, "rvifta")) != EOF) {
	switch (c) {
	case 'r':
	    opt_recursive = True;
	    break;
	case 'v':
	    opt_verbose = True;
	    break;
	case 'i':
	    opt_interactive = True;
	    break;
	case 'f':
	    opt_force = True;
	    break;
	case 't':
	    opt_test = True;
	    break;
	case 'a':
	    opt_all = True;
	    break;
	default:
	    usage();
	}
    }

    argc -= optind;
    argv += optind;

    if (argc < 1)
	usage();

    for (i = 0, pathes = 1, wilds = 0; i < argc; i++)
	if (check_path_or_file(argv[i])) {
	    strcpy(tmp, argv[i]);
	    if (!check_end_delimiter(tmp)) {
		if (*(tmp + 1) == ':' && *(tmp + 2) == '\0')
		    /* in order to get current directory (^^; */
		    strcat(tmp, ".");
		/*
		 * expand_wild ͈f~^ŏIĂƁA
		 * ƃfBNĝ݂}b`
		 */
		strcat(tmp, "/");
	    }
	    path_entry[pathes] = malloc_for_string(tmp);
	    strcpy(path_entry[pathes++], tmp);
	} else
	    parse_wild(argv[i], &wilds_entry[wilds++]);

    if (!wilds)
	usage();

    wilds_entry[wilds].wildmark = NULL;

    pathes = (pathes == 1 ? pathes + 1 : pathes);
    path_entry[pathes] = NULL;

    /* pXw肾fBNgȂ */
    _dosfind_param = 0x10 + 0x02 * opt_all;
    pathes = expand_wild(pathes, path_entry, &expanded_dirs);
    expanded_dirs++;		/* gȂ expanded_dirs[0] Ƃ΂ */
    for (i = 1; path_entry[i] != NULL; i++)
	free(path_entry[i]);

    _file_attr = 0x25 + 0x10 * opt_recursive + 0x02 * opt_all;

    /* -f -t -i ꍇ́Ax폜i\j֐ĂԁB */
    if (opt_force || opt_test || opt_interactive)
	r_exclusive(expanded_dirs, wilds_entry, opt_test);
    else
	/* mFpɈxtpX\邽߂ɁAxĂяoB */
	/* A߂̌ō폜ΏۂȂꍇ͔B */
    if (r_exclusive(expanded_dirs, wilds_entry, True)) {
	fprintf(stderr, "remove all ? ");
	if (YES_OR_NO_P())
	    r_exclusive(expanded_dirs, wilds_entry, False);
    }
    discard_argv(expanded_dirs);
    return error_code;
}
