 /* exit: 0=same, 1=diff, 2=error */
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include "tools.h"
#define RBUFSIZ 8192
#define SBUFSIZ 1024

#ifdef MSDOS
#  define RDMODE "rb"
#else
#  define RDMODE "r"
#endif

char	*cmdname = "cmp";
#ifdef LSI_C
static	FILE	*eofmsg_o = stderr;
#else /* some compilers' stderr is not constant. */
static	FILE	*eofmsg_o;
#endif

static	void	eofmsg(s)
char	*s;
{
	fprintf(eofmsg_o, "%s: EOF on %s\n", cmdname, s);
}
#define eofmsg_exit(s, i) (eofmsg(s), exit(i))

static	void	usage()
{
	fprintf(stderr,"Usage: %s [-lse] file1 file2 [skip1 skip2]\n",cmdname);
	exit(2);
}

static	void	cantbuf()
{
	fprintf(stderr, "%s: Can't set I/O buffer\n", cmdname), exit(2);
}

static	long	calc_amount(s)
char	*s;
{
	long	retval = 0L;

	switch(*s){
	case '0':
		switch(*++s){
		case 'x': case 'X':
			goto hexa;
		}
		 /* 8i */
		for(; is_octal(*s); ++s)
			(retval <<= 3, retval |= ctoi(*s));
		break;
	case 'x':
	case 'X':
	hexa: /* 16i */
		for(++s; is_xdigit(*s); ++s)
			(retval <<= 4, retval |= ctoi(*s));
		break;
	default:
		for(; is_digit(*s); s++)
			(retval *= 10L, retval += ctoi(*s));
		break;
	}
	return retval;
}

static	FILE	*argfopen(s, pass)
register char	*s, *pass;
{
	register FILE	*f;
	char	*malbuf;
	long	pass_amount;
	int	realbufsiz;

	if(*s == '-' && !s[1]){
		fsetbin(stdin);
		f = stdin;
	} else {
		if(NULL == (f = fopen(s, RDMODE))){
			fprintf(stderr, "%s: %s: Can't open\n", cmdname,
				conv_to_unix_format(s));
			exit(2);
		}
	}
	if(NULL == (malbuf = malloc(realbufsiz = RBUFSIZ)) &&
	   NULL == (malbuf = malloc(realbufsiz = SBUFSIZ))) cantbuf();
	if(NULL != pass){
		int	readonce, curread;

		pass_amount = calc_amount(pass);
		while(pass_amount){ /* 擪ǂݔ΂ */
			readonce = ((long)realbufsiz < pass_amount ?
					realbufsiz : (int)pass_amount);
			if(0 >= (curread = read(fileno(f), malbuf, readonce))){
#ifdef EINTR
				if(curread < 0 && errno == EINTR) continue;
#endif
				eofmsg_exit(s, 2); /* G[ */
			}
			pass_amount -= curread;
		}
	}
	if(0 != setfbuf(f, malbuf, realbufsiz)) cantbuf();
	return f;
}

#ifdef MSDOS
void	expand_hook()
{
	_expand_abendcode = 2;
	expand_hook_default();
}
#endif

static	int	s_cmp(s, p1, t, p2)
char	*s, *p1, *t, *p2;
{
	register FILE	*f = argfopen(s, p1), *g = argfopen(t, p2);
	register int	c;

	while(c = getc(f), c == getc(g)) if(c == EOF) return 0;
	return 1;
}

static	int	a_cmp(s, p1, t, p2)
char	*s, *p1, *t, *p2;
{
	FILE	*f = argfopen(s, p1), *g = argfopen(t, p2);
	char	status = 0;
	long	pos = 0;
	register int	c, d;

	while(c = getc(f), d = getc(g), ++pos, c != EOF && d != EOF){
		if(c == d) continue;
		status = 1;
		printf("%6ld %3o %3o\n", pos, c, d);
	}
	if(d != EOF){
		eofmsg(s); return 1;
	} else if(c != EOF){
		eofmsg(t); return 1;
	} else {
		return status;
	}
}

static	int	cmp(s, p1, t, p2)
char	*s, *p1, *t, *p2;
{
	FILE	*f = argfopen(s, p1), *g = argfopen(t, p2);
	long	pos = 0, line = 1;
	register int	c, d;

	for(; c = getc(f), d = getc(g), ++pos, c == d; ){
		switch(c){
		case EOF: return 0;
		case '\n': line++;
		}
	}
	if(c == EOF){
		eofmsg(s);
	} else if(d == EOF){
		eofmsg(t);
	} else {
		printf("%s %s differ: char %ld, line %ld\n", s, t, pos, line);
	}
	return 1;
}

main(ac, av)
int	ac;
char	*av[];
{
	extern	int	optind, opterr;
	extern	char	*optarg;
	int	c;
	bool	all = 0, silent = 0;
	char	*pass1 = NULL, *pass2 = NULL;

#ifdef EXPLICIT_EXPARG
	ac = expand_arg(av[0], get_psp_arg(), &av);
#endif
#ifndef MSDOS
	setcmdname(av[0]);
#endif
	opterr = 0;

#ifndef LSI_C
	eofmsg_o = stderr;
#endif
	while(EOF != (c = getopt(ac, av, "lsep:"))){
		cswitch(c){
		case 'l':
			all = 1; break;
		case 's':
			silent = 1; break;
		case 'e':
			eofmsg_o = stdout; break;
		case 'p': /* obsolete, undoc */
			pass1 = optarg, pass2 = strchr(pass1, ',');
			if(pass2 != NULL) ++pass2;
			break;
		default:
			usage(); break;
		}
	}

	ac -= optind, av += optind;
	switch(ac){
	case 4:
		pass2 = av[3]; /* through */
	case 3:
		pass1 = av[2]; /* through */
	case 2:
		break;
	default:
		usage();
	}

	return (silent? s_cmp : all? a_cmp : cmp)(av[0], pass1, av[1], pass2);
}
