/*
	Original by Yasushi Saito, TAIST.
	This program is NOT COPYRIGHTED,
	used, modified, re-distributed with no restriction,and no warranty.
*/

#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#ifndef UX_COMPILE_TEST
#  include <dos.h>
#  include <io.h>
#  if defined(__WATCOMC__) || defined(MSC51)
#    include <direct.h>
#  endif
#endif
#include <errno.h>
#include "tools.h"

#define YES_OR_NO_P yn_p_with_ret	/* y/nœƂ^[v */
/* #define YES_OR_NO_P yes_or_no_p */	/* y/nœƂ^[sv */

#ifdef SUPP_LFN
#  define BUFFERSIZE 1024 /* perhaps enough */
#else
#  define BUFFERSIZE 0x7fff /* max signed int */
#endif

char	*cmdname = "cp";

static	bool	opt_recursive, opt_verbose, opt_interactive, opt_backup,
		opt_preserve, opt_update, opt_presv_sfn, opt_visonly,
		opt_setarcbit, opt_force;
/* initially 0 */

static	int	error_code;

#define FATALEXIT 2
#ifdef MSDOS
void	expand_hook()
{
	_expand_abendcode = FATALEXIT;
	expand_hook_default();
}
#endif

static	void	cp_error(s, reason)
char	*s, *reason;
{
	if(opt_force) return;
	fputss(stderr, cmdname, ": ", s_conv_to_unix_format(s), ": ",
		reason, ".\n", NULL);
	error_code = 1;
}

static	void	io_error(msg, s)
char	*msg, *s;
{
	fputss(stderr, cmdname, ": ", msg, s_conv_to_unix_format(s), ": ",
		strerror(errno), ".\n", NULL);
	if(opt_force) return;
	error_code = 1;
}

static	char	copybuf[BUFFERSIZE];

static	void	verbose_out(source, dest)
char	*source, *dest;
{
	fputs(s_conv_to_unix_format(source), stderr);
	fputss(stderr, " -> ", s_conv_to_unix_format(dest), "\n", NULL);
}

struct filetime {
#ifdef SUPP_LFN
	WREGTYPE ad; /* access date (time not supported)
			  L^ĂȂꍇmodiry dateƓ̂ */
#endif
	WREGTYPE d, t;
};

/* Do copying 'source' to 'dest'. */
 /* fdfile handlepĂ̂̓IWiォB܂Ȃ? Nide */
static	void	copy_file(source, dest)
char	*source, *dest;
{
	int	fds, fdd, n;
	unsigned dest_stat;
	struct	filetime src_dt;

	if((unsigned)-1 == get_file_status(source)){ /* DOS append΍ */
		struct	Find_t	tmp;

		if(find_Fixedfile(source, &tmp, 0) ||
		   !((char)tmp . attrib & _A_DEV)){
			cp_error(source, strerror(ENOENT));
			return;
		} /* cp con file ̂悤ȃP[X͋ */
	}

#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
	if(_lfnsupp_level > LFN_NONE)
		get_amtime(source, &src_dt . d, &src_dt . t, &src_dt . ad);
		 /* get_amtime(), set_amtime()LFNT|[ggĂ
		    ȂȂB */
	else
		src_dt . d = 0;
#endif
	if((fds = open(source, O_RDONLY | O_BINARY)) < 0){
		io_error("Can't open source ", source);
		return;
	}
#ifdef SUPP_LFN
	if(!src_dt . d)
#endif
	{
		_dos_getftime(fds, &src_dt . d, &src_dt . t);
#ifdef SUPP_LFN
		src_dt . ad = src_dt . d;
#endif
	}

	dest_stat = get_file_status(dest);
	if(opt_update && dest_stat != (unsigned)-1 && /* DOSappend΍ */
	   0 <= (fdd = open(dest, O_RDONLY | O_BINARY))){
		struct	filetime dest_dt;

		_dos_getftime(fdd, &dest_dt . d, &dest_dt . t);
		close(fdd);
		if(src_dt . d < dest_dt . d ||
		   src_dt . d == dest_dt . d && src_dt . t <= dest_dt . t){
#if 0
			fprintf(stderr, "%s: older than ",
				s_conv_to_unix_format(source));
			fprintf(stderr, "%s; not copied\n",
				s_conv_to_unix_format(dest));
#endif
			close(fds);
			return;
		}
	}
	if(opt_verbose) verbose_out(source, dest);

	if(dest_stat != (unsigned)-1 && (opt_interactive || opt_backup)){
		if(opt_backup){
			if(opt_verbose)	fputss(stderr, "    (backup ",
				s_conv_to_unix_format(dest), ")\n", NULL);
			if(!rename_as_backup(dest)){
				io_error("Can't backup ", dest);
				goto ask;
			}
			dest_stat = -1;
		} else {
		ask:
			fputss(stderr, s_conv_to_unix_format(dest),
				" exists, overwrite? ", NULL);
			if(!YES_OR_NO_P()){
				close(fds);
				return;
			}
		}
	}
	for(n = 0; n < 2; n++){
		if((fdd = open(dest,
		    (dest_stat != (unsigned)-1 ? /* DOSappend΍ */
			O_TRUNC|O_WRONLY|O_BINARY :
			O_CREAT|O_EXCL|O_WRONLY|O_BINARY),
		    0644)) >= 0) break;

		if(opt_force && !n && errno == EACCES &&
		   !(dest_stat & _A_SUBDIR) &&
			/* SUBDIRȂ̂EACCESɂȂ鎞񂾁c */
		   !_dos_setfileattr(dest, _A_NORMAL)) continue;
		io_error("Can't open destination ", dest);
		close(fds);
		return;
	}
	while((n = read(fds, copybuf, BUFFERSIZE)) > 0){
		if(write(fdd, copybuf, n) != n){
			io_error("Write error: ", dest);
			goto error;
		}
	}
	if(n < 0){ /* add Nide, mvɂ͂̏̂cpɂt */
		io_error("Read error: ", source);
	error:
		close(fds);
		close(fdd);
		unlink(dest);
		exit(1);
	}

	if(opt_preserve){
		/* set_modification_time(fds, fdd); */
		_dos_setftime(fdd, src_dt . d, src_dt . t);
	}
	close(fdd);
	close(fds);
#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
	if(_lfnsupp_level > LFN_NONE && opt_preserve){
		set_amtime(dest, src_dt . d, src_dt . t, src_dt . ad);
	}
#endif

	if(opt_preserve || dest_stat == (unsigned)-1){
		copy_attribute_with_mask(source, dest,
						opt_setarcbit ? _A_ARCH : 0);
	} else {
		if(opt_setarcbit) dest_stat |= _A_ARCH;
		_dos_setfileattr(dest, dest_stat & 0x27);
	} /* Check: Is this same as UNIX? (Nide) *//* also check usage */
}

/* Copy 'source' to 'dest'. Copy recursively if 'source' is a directory and
   'recursive' option is set. */
static	void	r_copy_file(source, dest)
char	*source, *dest;
{
	int	modes;

	/* Strip off last \\ in the filename string */
	modes = get_file_status(source);
	conv_to_dos_regpath(dest);

	/* Check if source and destination are same, or source path is the
	   part of the destination(recursive copy). */
	if(chk_recursive(source, dest)){
		cp_error(source,
	  "Seems to be identical to the destination file, or recursive copy");
		return;
	}

	if(!opt_recursive){
		if(modes != (us_int)-1 && (modes & _A_SUBDIR)){
			cp_error(source,
			 "Directory is not copied unless -r option specified");
		} else
#if 0
		if(conv_to_dos_regpath(dest)){ /* PATH ends with \ */
			cp_error(dest, "Is a directory");
		} else
#endif
			copy_file(source, dest);
			 /* source񑶍݂̏ꍇ͌Ń`FbN */
	} else {
		struct	Find_t	*f;
		char	*temp, *tempd;
		unsigned destattr;
		bool	empty_root = 0;

		if(modes != (us_int)-1 && (modes & _A_SUBDIR) &&
		   !(get_file_status(source) & _A_SUBDIR)){
		 /* sourcefBNgłȂ̂ɍŌɁu/vĂ邹
		    get_file_status(source)fBNgƔfꍇ
		    ̂ōĊmFBx͍Ṓu/v͏ĂB
		    source񑶍݂̏ꍇ͌Ń`FbN */
			cp_error(source, "Not a directory");
			return;
		}

		 /* drecursivẽX^bNsh~̂autozɂȂ */
		f = chk_malloc(sizeof(struct Find_t) + PATHNAMELENGTH * 2);
		tempd = (temp = ((char *)f) + sizeof(struct Find_t)) +
			PATHNAMELENGTH;

		expand_path(temp, source, ALLNAME);
		if(_dos_Findfirst(temp, opt_visonly ? 0x31 : 0x37, f)){
			int	tmpfd;

			expand_path(temp, source, NULNAME);
			if(0 <= (tmpfd = open(temp, O_RDONLY))){
			 /* :maybe empty root */
				close(tmpfd);
				empty_root = 1;
				goto child_found;
			}
			copy_file(source, dest);
		} else {
		child_found:
			if(opt_verbose) verbose_out(source, dest);

			if(_dos_getfileattr(dest, &destattr) ?
#ifdef UX_COMPILE_TEST
			     mkdir(dest, 0755) :
#else
			     mkdir(dest) :
#endif
			     !(destattr & _A_SUBDIR)){
				cp_error(dest, "Can't make directory");
			} else {
#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
 /* Win95łdiratime̓t@C̍쐬폜łςȂ݂ */
				if(_lfnsupp_level > LFN_NONE && opt_preserve){
					struct	filetime sdir_dt;

					if(!get_amtime(source, &sdir_dt . d,
					    &sdir_dt . t, &sdir_dt . ad)){
						set_amtime(dest, sdir_dt . d,
						    sdir_dt . t, sdir_dt . ad);
					}
				}
#endif
				if(!empty_root) do {
					char	*name;

					 /* If the first character of the
					    filename is '.', then the filename
					    must be "." or ".." in MSDOS. */
					if(
#ifdef SUPP_LFN
					   name = f -> lname,
					   is_self_or_parent_dir(name)
#else
					   *(name = f -> name) == '.'
#endif
					) continue;

					expand_path(temp, source, name);
					expand_path(tempd, dest, name);
					(((char)f -> attrib & _A_SUBDIR) ?
					 r_copy_file : copy_file)(temp, tempd);
				} while(!_dos_Findnext(f));
				copy_attribute(source, dest);
			}
			if(!empty_root)	_dos_Findclose(f);
		}
		free(f);
	}
}

static	void	multifile_copy(filenum, filevec)
int	filenum;
char	**filevec;
{
	char	buf[PATHNAMELENGTH], *p, *q;
#ifdef SUPP_LFN
	char	lfntmp[PATHNAMELENGTH];
#endif
	int	i;

	for(i = 0; i < filenum - 1; ++i){
		q = filename_base(p = filevec[i]);

		expand_path(buf, filevec[filenum - 1],
			is_ancestor_dir(q) ?
				NULSTR :
#ifdef SUPP_LFN
			_lfnsupp_level > LFN_NONE && !opt_presv_sfn &&
			!abs_pathname(p, lfntmp, 2) ?
				filename_base(lfntmp) :
#endif
				q
		);
		r_copy_file(p, buf);
	}
}

static	void	help_and_exit()
 /* Nide, ANSIRpC̕ւ̂߂ɕAڂ߂ */
{
	fputss(stderr, "Usage:\t",
		cmdname, " [-rviabpfusx] file1 ... filen directory\n\t",
		cmdname, " [-rviabpfusx] file1 file2\n", NULL);
	exit(FATALEXIT);
}

main(ac, av)
int	ac;
char	*av[];
{
	int	c;
	unsigned mode;
	extern	int	optind, opterr;
	extern	char	*optarg;

#ifdef EXPLICIT_EXPARG
	ac = expand_arg(av[0], get_psp_arg(), &av);
#endif
#ifndef MSDOS
	setcmdname(av[0]);
#endif
	opterr = 0;

	while((c = getopt(ac, av, "rviabpfusx")) != EOF){
		cswitch(c){
		case 'r':
			opt_recursive = TRUE;
			break;
		case 'v':
			opt_verbose = TRUE;
			break;
		case 'i':
			opt_interactive = TRUE;
			break;
		case 'a':
			opt_visonly = TRUE;
			break;
		case 'b':
			opt_backup = TRUE;
			break;
		case 'p':
			opt_preserve = TRUE;
			break;
			/* I use "alias cp cp -p" :-) */
		case 'f':
			opt_force = TRUE;
			break;
		case 'u':
			opt_update = TRUE;
			break;
		case 's':
			opt_presv_sfn = TRUE;
			break;
		case 'x':
			opt_setarcbit = TRUE;
			break;
		default:
			help_and_exit();
		}
	}

	av += optind, ac -= optind;
	if(ac < 2) help_and_exit();

	if((mode = get_file_status(av[ac - 1])) != (us_int)-1 &&
	   (mode & _A_SUBDIR)){
		/* The last argument is intended or found to be a directory */
		multifile_copy(ac, av);
	} else {
		if(ac >= 3) help_and_exit();
		r_copy_file(av[0], av[1]);
	}
	return error_code;
}
