#include <stdio.h>
#include <ctype.h>
#include "tools.h"
#ifndef MSDOS
#  define EUCJMODE
#  define iskanji(c) ((c) & 0x80) /* ꉞ */
#endif

#ifdef EUCJMODE
#  define KSPC	0xa1a1
#  define SS2	0x8e
#else
#  define KSPC	0x8140
#endif

#define	COLLISTMAX 512

char	*cmdname = "expand";

static	int	tabcol = 8;
static	int	tabcols[COLLISTMAX];
static	bool	addsp = 0, eofnl = 0, kspc = 0, ionly = 0;

static	void	usage()
{
	_puterrmes("Usage: "), _puterrmes(cmdname);
	_puterrmes_cr(" [-nski] [-tablist] [file...]");
	exit(1);
}

static	void	do_opt(opt)
char	*opt;
{
	int	tabidx = 0, prevcol = 0;

	for(;;){
		if(tabidx == COLLISTMAX - 1){
			_cmdputerrmes_cr("Tabstop list too long");
			exit(1);
		}
		tabcols[tabidx] = atoi(opt);
		if(prevcol >= tabcols[tabidx]){
			_cmdputerrmes_cr("Bad tab stop spec");
			exit(1);
		} else prevcol = tabcols[tabidx++];
		while(is_digit(*opt)) ++opt;
		if(!*opt) break;
		if(*opt++ != ',') usage();
	}
	tabcols[tabidx] = 0;
	tabcol = (tabcols[1] ? 0 : tabcols[0]);
}

static	expand(f) /* returns last pos */
FILE	*f;
{
	int	pos = 0, c, d, tabidx = 0;
	bool	non_init = 0;

	while(EOF != (c = getc(f))){
		if(iskanji(c)){
			if(EOF == (d = getc(f))){
				putchar(c);
				return ++pos;
			}
			if(kspc && ((c << 8) | d) == KSPC) c = d = ' ';
			putchar(c), putchar(d);
#ifdef EUCJMODE
			++pos;
			if(c != SS2) ++pos;
#else
			pos += 2;
#endif
			non_init = 1;
		} else
		if(c == '\b'){
			putchar(c);
			if(pos) --pos;
			if(tabidx) --tabidx; /* ō1߂΂悢 */
		} else
		if(c == '\n' || c == '\r' || c == '\f'){
			if(!pos && addsp) putchar(' ');
			putchar(c), pos = tabidx = 0, non_init = 0;
		} else
		if(c == '\t'){
			int	newpos;

			if(tabcol){
				newpos = (pos / tabcol + 1) * tabcol;
			} else {
				for(;; ++tabidx){
					newpos = tabcols[tabidx];
					if(!newpos || pos < newpos) break;
				}
				if(!newpos) newpos = pos + 1;
			}
			if(ionly && non_init){
				putchar('\t'), pos = newpos;
			} else {
				for(; pos < newpos; ++pos) putchar(' ');
			}
		} else {
			putchar(c), ++pos;
			if(c != ' ') non_init = 1;
		}
	}
	return pos;
}

main(ac, av)
int	ac;
char	**av;
{
	FILE	*inf;
	int	pos;

#ifdef EXPLICIT_EXPARG
	ac = expand_arg(av[0], get_psp_arg(), &av);
#endif
#ifndef MSDOS
	setcmdname(av[0]);
#endif

	--ac, ++av;
	while(ac && av[0][0] == '-' && av[0][1]){
		char	*p;

		for(p = *av + 1; *p && !is_digit(*p); ++p){
			switch(*p){
			case 's':
				addsp = 1; break;
			case 'k':
				kspc = 1; break;
			case 'n':
				eofnl = 1; break;
			case 'i':
				ionly = 1; break;
			default:
				usage();
			}
		}
		if(*p) do_opt(p);
		--ac, ++av;
	}
	if(!ac){
		pos = expand(stdin);
	} else {
		for(; ac; --ac, ++av){
			if(av[0][0] == '-' && !av[0][1]){
				pos = expand(stdin);
				clearerr(stdin);
			} else {
				if(NULL == (inf = fopen(*av, "r"))){
					_cmdputerrmes(
						s_conv_to_unix_format(*av));
					_puterrmes_cr(": Can't open");
					exit(1);
				}
				pos = expand(inf);
				fclose(inf);
			}
		}
	}
	if(pos && eofnl) putchar('\n');
	return 0;
}
