/*
	Original by Yasushi Saito, TAIST.
	This program is in public domain.
	Used, modified, re-distributed with no restriction, and no warranty.
*/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#ifndef UX_COMPILE_TEST
#  include <dos.h>
#  include <direct.h>
#  include <io.h>
#endif
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "tools.h"

#ifdef SUPP_LFN
#  define BUFFERSIZE 1024 /* perhaps enough */
#else
#  define BUFFERSIZE 0x7fff /* max signed int */
#endif

#define YES_OR_NO_P yn_p_with_ret	/* y/nœƂ^[v */
/* #define YES_OR_NO_P yes_or_no_p */	/* y/nœƂ^[sv */

char	*cmdname = "mv";

static	bool	opt_force, opt_verbose, opt_interactive, opt_backup,
		opt_update, opt_presv_sfn; /* initially 0 */

#define FATALEXIT 2
#ifdef MSDOS
void	expand_hook()
{
	_expand_abendcode = FATALEXIT;
	expand_hook_default();
}
#endif

/* Error reporting. */
static	error_code = 0;

static	void	mv_error(s, reason)
char	*s, *reason;
{
	if(opt_force) return;
	if(s != NULL) fputss(stderr, cmdname, ": ",
		s_conv_to_unix_format(s), ": ", NULL);
	fputss(stderr, reason, ".\n", NULL);
	error_code = 1;
}

static	void	io_error(msg, s)
char	*msg, *s;
{
	if(msg != NULL) fputss(stderr, cmdname, ": ", msg, NULL);
	fputss(stderr, s_conv_to_unix_format(s), ": ",
		strerror(errno), ".\n", NULL);
	if(opt_force) return;
	error_code = 1;
}

static	char	copybuf[BUFFERSIZE];

static	void	verbose_out(source, dest)
char	*source, *dest;
{
	 /* s_conv_to_unix_format()2xĂԂƐ̌ʂ󂷂̂
	    2ɕďo */
	fputs(s_conv_to_unix_format(source), stderr);
	fputss(stderr, " -> ", s_conv_to_unix_format(dest), "\n", NULL);
}

struct filetime {
#ifdef SUPP_LFN
	WREGTYPE ad; /* access date (time not supported)
			  L^ĂȂꍇmodiry dateƓ̂ */
#endif
	WREGTYPE d, t;
};

static	un_readonly_chk(mode, fnm, flg)
unsigned mode, flg;
char	*fnm;
 /* flg bit:0...is the destination, bit:1...no actual modify
    return -1...answered `no' (then doesn't perform moving)
	    0...was not readonly
	    1...attr changed */
{
	if(mode == (us_int)-1 || !(mode & _A_RDONLY)) return 0;

	if(opt_interactive || !opt_force){
		fputss(stderr, s_conv_to_unix_format(fnm),
			(flg & 1) && opt_interactive ? " exists and" : NULSTR,
			" readonly, move? ", NULL);
		if(!YES_OR_NO_P()) return -1;
	}
	if(!(flg & 2)) _dos_setfileattr(fnm, mode & ~7);
	return 1;
}

static	int	rename_and_errchk(source, dest)
char	*source, *dest;
{
	if(!nonshare_chk_rename(source, dest)) return 0;

	fputss(stderr, cmdname, ": Couldn't move ",
		s_conv_to_unix_format(source), " to ", NULL);
	io_error(NULL, dest);
	return 1;
}

static	int	unlink_and_errchk(dest)
char	*dest;
{
	if(!nonshare_chk_unlink(dest)) return 0;

	io_error("Can't delete ", dest);
	return 1;
}

static	void	move_file();

 /* move directory. this routine is moved from within move_file() */
static	void	move_dir(source, dest)
char	*source, *dest;
{
	struct	Find_t	*f;
	char	*s, *ss, *dd;

	if(!rename(source, dest)) return;
	 /* func56h7156hɒuĂꍇfBNgrename
	    邱Ƃ(Ƃ肻DOS7ł56hłfBNg
	    rename?) */

#ifndef UX_COMPILE_TEST
	if(mkdir(dest))
#else
	if(mkdir(dest, 0755))
#endif
	{
		/* Failed on making destination directory */
		mv_error(dest, "Can't make directory");

		return;
	}

#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
 /* Win95łdiratime̓t@C̍쐬폜łςȂ݂ */
	if(_lfnsupp_level > LFN_NONE){
		struct	filetime sdir_dt;

		if(!get_amtime(source, &sdir_dt.d, &sdir_dt.t, &sdir_dt.ad)){
			set_amtime(dest, sdir_dt.d, sdir_dt.t, sdir_dt.ad);
		}
		 /* get_amtime(), set_amtime()LFNT|[ggĂ
		    ȂȂB */
	}
#endif

	f = chk_malloc(sizeof(struct Find_t) + PATHNAMELENGTH * 3);
	 /* drecursivẽX^bNsh~̂autozɂȂ */
	s = ((char *)f) + sizeof(struct Find_t);
	dd = (ss = s + PATHNAMELENGTH) + PATHNAMELENGTH;

	/* Read the filenames contained in the source directory and
	   recursively move each files. */
	expand_path(s, source, ALLNAME);
	if(_dos_Findfirst(s, 0x16, f)){
		int	tmpfd;

		expand_path(s, source, NULNAME);
		if(0 <= (tmpfd = open(s, O_RDONLY))){
		 /* :maybe empty root */
			close(tmpfd);
			goto copy_dirattr;
		}
		mv_error(source, "Can't read directory");

	} else {
		char	*name;

		do {
			if(
#ifdef SUPP_LFN
			   name = f -> lname, is_self_or_parent_dir(name)
#else
			   *(name = f -> name) == '.'
#endif
			) continue;

			expand_path(ss, source, name);
			expand_path(dd, dest, name);
			move_file(ss, dd);
		} while(!_dos_Findnext(f));
		_dos_Findclose(f);
	copy_dirattr:	/* Add Nide, dir attr copy */
		copy_attribute(source, dest);
		if(rmdir(source))
			mv_error(source, "Can't delete directory");
	}
	free(f);
}

static	int	dest_exchk_byopts(dest, moded, moded_changed)
char	*dest;
us_int	moded;
int	moded_changed;
 /* 0:dest񑶍 1:backupς 2:݂邪mv邩?N 3:destmv */
{
	if(moded == (unsigned)-1) return 0;

	if(opt_backup){
		if(opt_verbose)	fputss(stderr, "    (backup ",
				s_conv_to_unix_format(dest), ")\n", NULL);
		if(!ren_as_bkup_set_attr(dest, moded)){
			io_error("Can't backup ", dest);
			goto ask;
		}
		return 1;
	} else {
		if(opt_interactive && !moded_changed){
		 /* moded_changedȂ炱ɂ͗ȂB̏ꍇ͊
		    udest݂readonlymv邩?vƖ₢킹ς */
		ask:
			fputss(stderr,	s_conv_to_unix_format(dest),
						" exists, move? ", NULL);
			if(!YES_OR_NO_P()) return 2;
		}
		return 3;
	}
}

/* Move SOURCE to DEST. */
 /* fdfile handlepĂ̂̓IWiォB܂Ȃ? Nide */
static	void	move_file(source, dest)
char	*source, *dest;
{
	int	fds, fdd;	/* File descriptor for each file. */
	int	disks, diskd;	/* Device name for each file */
	unsigned modes, moded;	/* File attribute for each file */
	struct	filetime dates;	/* File time for source */
	int	n;		/* Counter when copying */
	int	modes_changed = 0, moded_changed = 0;
				/* File attr changed flag */

	modes = get_file_status(source);
	moded = get_file_status(dest);

	/* Check the attribute of each file and do pre-processing. */
	if(modes == (us_int)-1){
	nosrc:
		mv_error(source, strerror(ENOENT));
		return;
	}
	/* Check if two files are the same */
	if(chk_recursive(source, dest)){
		mv_error(source,
	  "Seems to be identical to the destination file, or recursive move");
		return;
	}

	if(modes & _A_SUBDIR){
		 /* sourcefBNgłȂ̂ɍŌɁu/vĂ邹
		    get_file_status(source)fBNgƔfꍇ
		    ̂ōĊmFBx͍Ṓu/v͏Ă */
		if((modes = get_file_status(source)) == (us_int)-1)
			goto nosrc;
		if(!(modes & _A_SUBDIR)){
			mv_error(source, "Not a directory");
			return;
		}

		if(moded != (unsigned)-1){
			/* Does not move when destination already exists and
			   is a plain file */
			fputss(stderr, cmdname, ": destination ",
				s_conv_to_unix_format(dest), ": ", NULL);
			mv_error(NULL,
				"Can't move directory to existent file");
			return;
		}
		if(opt_verbose) verbose_out(source, dest);
		move_dir(source, dest);
		return;
	}

#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
	if(_lfnsupp_level > LFN_NONE)
		get_amtime(source, &dates . d, &dates . t, &dates . ad);
	else
		dates . d = 0;
#endif
	if((fds = open(source, O_RDONLY | O_BINARY)) < 0){
		io_error("Can't open source ", source);
		return;
	} /* fds is opened! if return due to error, close it */
#ifdef SUPP_LFN
	if(!dates . d)
#endif
	{
		_dos_getftime(fds, &dates . d, &dates . t);
#ifdef SUPP_LFN
		dates . ad = dates . d;
#endif
	}

	if(opt_update && moded != (unsigned)-1 && /* DOSappend΍ */
	   0 <= (fdd = open(dest, O_RDONLY | O_BINARY))){
		struct	filetime dated;	/* File time for dest */

		_dos_getftime(fdd, &dated . d, &dated . t);
		close(fdd);
		if(dates . d < dated . d || /* dates:source, dated:dest */
		   dates . d == dated . d && dates . t <= dated . t){
#if 0
			fprintf(stderr, "%s: older than ",
				s_conv_to_unix_format(source));
			fprintf(stderr, "%s; not moved\n",
				s_conv_to_unix_format(dest));
#endif
			close(fds);
			return;
		}
	}
	if(opt_verbose) verbose_out(source, dest);

	if(0 > (modes_changed = un_readonly_chk(modes, source, 0)) ||
	   0 > (moded_changed = un_readonly_chk(moded, dest, 1))){
	straighten_up_and_return2:
		close(fds);
		 /* attributes changed! if return due to error, recover them */
	straighten_up_and_return:
		if(modes_changed > 0) _dos_setfileattr(source, modes);
		if(moded_changed > 0) _dos_setfileattr(dest, moded);
		return;
	}

	 /* Check if we only have to change the 'link' of the file, or we have
	    to copy the contents to the destination and unlink the source. */
	disks = getfullpath_drv(NULL, source, -1);
#ifdef SUPP_LFN
	diskd = ancestor_pathdrv(NULL, dest, -1);
	 /* sourcȇ݂͂킩ĂBdest́A݂ł߂c悪
	    sourceƓhCuɂ邩 */
#else
	diskd = getfullpath_drv(NULL, dest, -1);
#endif

	 /* now fds is opened */
	if(disks >= 0 && disks == diskd){
		close(fds);
		/* We only have to change the 'link' of files if two files
		   are on the save device */
		switch(dest_exchk_byopts(dest, moded, moded_changed)){
		case 1:
			moded = -1;
			break;
		case 3:
			if(!unlink_and_errchk(dest)) break;
			 /* FALLTHROUGH */
		case 2:
			goto straighten_up_and_return;
		}

		if(rename_and_errchk(source, dest))
			goto straighten_up_and_return;
		_dos_setfileattr(dest, modes);	/* Add Nide; copy attribute */
	} else {
		/* Copy, and unlink , when two files are on different
		   devices. */
		switch(dest_exchk_byopts(dest, moded, moded_changed)){
		case 1:
			moded = -1;
			break;
		case 2:
			goto straighten_up_and_return2;
		}

#if 0		 /* In previous versions fds is opened here */
		if((fds = open(source, O_RDONLY | O_BINARY)) < 0){
			io_error("Can't open source ", source);
			return;
		}
#endif
		if((fdd = open(dest, (moded != (unsigned)-1 ?
		   O_TRUNC|O_WRONLY|O_BINARY:O_CREAT|O_EXCL|O_WRONLY|O_BINARY),
		   0644)) < 0){ /* modeďDOSappend΍ */
			io_error("Can't open destination ", dest);
			goto straighten_up_and_return2;
		}

		while((n = read(fds, copybuf, BUFFERSIZE)) > 0){
			if(write(fdd, copybuf, n) != n){
				io_error("Write error: ", dest);
				goto error;
			}
		}
		if(n < 0){
			io_error("Read error: ", source);
		error:	close(fdd);
			close(fds);
			unlink(dest), _dos_setfileattr(source, modes);
			exit(1);
		}

		_dos_setftime(fdd, dates . d, dates . t);
		/* set_modification_time(fds, fdd); */
		close(fdd);
		close(fds);
#if defined(SUPP_LFN) && !defined(UX_COMPILE_TEST)
		if(_lfnsupp_level > LFN_NONE){ /* mvł͖Ɏԕۑ */
			set_amtime(dest, dates . d, dates . t, dates . ad);
		}
#endif

		_dos_setfileattr(dest, modes);	/* Add Nide; copy attribute */
		unlink_and_errchk(source);
	}
}

static	void	multifile_move(filenum, filevec)
int	filenum;
char	**filevec;
{
	char	buf[PATHNAMELENGTH], *p, *q;
#ifdef SUPP_LFN
	char	lfntmp[PATHNAMELENGTH];
#endif
	int	i;

	for(i = 0; i < filenum - 1; ++i){
		q = filename_base(p = filevec[i]);

		expand_path(buf, filevec[filenum - 1],
			is_ancestor_dir(q) ?
				NULSTR :
#ifdef SUPP_LFN
			_lfnsupp_level > LFN_NONE && !opt_presv_sfn &&
			!abs_pathname(p, lfntmp, 2) ?
				filename_base(lfntmp) :
#endif
				q
		);
		*(last_pathsep(buf, NULL)) = '\0';
		 /* buf̍Ṓu/vĂBmv dir/ destdir ̎
		    bufudestdir/dir/v̂܂܂destdirȂĂ
		    move_filemoded0x10ɂȂāucan't move to
		    existent fileṽG[N */
		move_file(p, buf);
	}
}

static	void	ren_case(source, dest)
char	*source, *dest;
{
	us_int	modes = get_file_status(source);

	if(!(modes & _A_SUBDIR) && 0 > un_readonly_chk(modes, source, 2))
		return;
	 /* _ł̓fBNgmvreadonly𖳎Ă̂ŁA
	    킹Ă */
	rename_and_errchk(source, dest);
}

static	void	help_and_exit()
/* Nide, ANSIRpC̕ւ̂߂ɕAڂ߂ */
{
	fputss(stderr, "Usage:\t",
		cmdname, " [-vibfus] file1 ... filen directory\n\t",
		cmdname, " [-vibfus] file1 file2\n", NULL);
	exit(FATALEXIT);
}

main(ac, av)
int	ac;
char	*av[];
{
	int	c;
	unsigned mode;
	extern	int	optind, opterr;
	extern	char	*optarg;

#ifdef EXPLICIT_EXPARG
	ac = expand_arg(av[0], get_psp_arg(), &av);
#endif
#ifndef MSDOS
	setcmdname(av[0]);
#endif
	opterr = 0;
	while((c = getopt(ac, av, "vibfus")) != EOF){
		cswitch(c){
		case 'f':
			opt_force = TRUE;
			break;
		case 'v':
			opt_verbose = TRUE;
			break;
		case 'i':
			opt_interactive = TRUE;
			break;
		case 'b':
			opt_backup = TRUE;
			break;
		case 'u':
			opt_update = TRUE;
			break;
		case 's':
			opt_presv_sfn = TRUE;
			break;
		default:
			help_and_exit();
		}
	}

	av += optind, ac -= optind;
	if(ac < 2) help_and_exit();

#ifdef SUPP_LFN
	if(ac == 2 && _lfnsupp_level > LFN_NONE &&
	   0 > chk_recursive(av[0], av[1])){
		ren_case(av[0], av[1]); /* av[1]n΂悢?:-) */
	} else
#endif
	if((mode = get_file_status(av[ac - 1])) != (us_int)-1 &&
	   (mode & _A_SUBDIR)){
		/* The last argument is intended or found to be a directory */
		multifile_move(ac, av);
	} else {
		if(ac >= 3) help_and_exit();
		move_file(av[0], av[1]);
	}
	return error_code;
}
