/*
 * Regular expression pattern matching routines, from comp.sources.misc
 * For a description of what this software does, read regex.c
 */
-- libr header start.
#include <stdio.h>
/* #include <string.h> moved to regexp.h */
#include <ctype.h>
#include "regexp.h"
/* #include "regmagic.h" */

	/* -------------------- from regmagic.h -------------------- */
/*
 * The first byte of the regexp internal "program" is actually this magic
 * number; the start node begins in the second byte.
 */
#define	MAGIC	0234
-regex.c
/*
  Author: Tasos Kotsikonas (tasos@cs.bu.edu)
  Disclaimer: All proper disclaimers apply; these include death, damage of any
	      kind to anything and anyone, and this kind of legal garbage.
  Copying: You may copy, alter and redistribute this software as you see fit,
	   but you may not sell it for profit.

  Do pattern matching of two flavors: strict egrep(1), or extended egrep(1).
  If the symbol egrep is #define'd below, strict egrep(1) syntax is used.

  The main routine is re_strcmp() which takes a regular expression, a
  string to match against, and a string specifying which of the matched
  subparts are to be used to form a new string -- if this string is NULL no
  such action is taken.

  The new string is stored in the same space, so this string should be
  long enough not to cause memory overwrites. To be able to do this match
  substitution, the desired subparts of the regular expression should
  be enclosed in ( and ). Each subpart is referenced to as \n where n is
  a digit from 1 to 9.

  For example, the following call:

  strcpy (match, "\\2@\\1.UUCP");
  re_strcmp ("[^!@]*!([^!@.]*)!([^!@]*)@.*", "GATE!HOP!USER@UUCP.SOME.COM",
	     match);

  will return TRUE and store in 'match' "USER@HOP.UUCP". See the man page
  for Henry Spencer's routines included for more information.

  In extended mode, regular expressions may use the logical operators ~ (not),
  & (and) and group expressions further with < and >. Because of the latter,
  < and > may not be used in the ed(1) sense (\< is the default in the
  context these routines were developed, and \> is useless in the same context).
  To use any of these characters literally, precede them with a backslash (\).

  For example:

  re_strcmp ("~<.*\.COM|.*\.EDU>&~TASOS*", "TASOS@FOO.US", NULL)

  will return FALSE.

  re_strcmp() returns TRUE on success, FALSE if no match was found, or
  -1 on error. In the latter case, the (char *) variable regerr contains the
  actual error message.

  The system uses modified pattern matching routines written by Henry Spencer
  (Copyright (c) 1986 by University of Toronto) -- actually only regerror.c
  was modified. Many thanks and kudos to Henry for this wonderful piece of code.

  It is possible to use the ls(1) style wild characters * and ? by altering
  the symbols STAR and QMARK below.

  It is also possible to turn off ed(1) style pattern matching (i.e. turn
  off the meaning of [ ] { } etc.) by using the function escape_re() where
  marked with ###.

  To test these routines compile with -Dtest and run independently as follows:

		% cc -Dtest -I. regexp.c regsub.c regex.c regerror.c
		% a.out 'regular-expression' file
*/

/* Define it if you want to use strict egrep(1) regular expressions */
#define egrep

#define noreport /* Add Nide */

#define MAXLENGTH	1024	/* Maximum length of each regular expression */
#define EOS		'\0'

#ifdef test
int  main();
#else
#ifndef noreport /* Add Nide */
extern void report_progress();
extern FILE *report;
#endif
#endif
#ifndef egrep
void escape_re();
char *convert_re();
int  icp();
int  push();
int  pop();
int  do_op();
void pop_op();
int  new_op();
int  eval();
int  isop();
int  prevch();
int  nextch();

#define STAR		"*"	/* If ".*" then the meaning is that of ls(1) */
#define QMARK		"?"	/* If "." then the meaning is that of ls(1) */
#define LGROUPCH	'<'	/* Avoid ( and ) */
#define RGROUPCH	'>'
#define OR		1	/* Operators should be > 0 */
#define AND		2
#define	NOT		3
#define LPAREN		4
#define RPAREN		5

typedef struct _operator_stack {
  int op;
  int isp;	/* In-stack priority */
  struct _operator_stack *next, *prev;
} OPERATOR_STACK;

typedef struct _operand_stack {
  int val;
  struct _operand_stack *next, *prev;
} OPERAND_STACK;

OPERATOR_STACK *op_top;
OPERAND_STACK *val_top;
int pliteral, literal, nliteral;
#endif

extern char *regerr;

#ifdef test
int	main(argc, argv)
int	argc;
char	**argv;
{
  FILE *f;
  char s [1024], matches [1024];
  int match;

  if (argc < 2)
    printf ("Usage: %s 'regular-expression' file\n", argv[0]), exit (1);
  if ((f = fopen (argv[2], "r")) == NULL)
    printf ("%s: ", argv[2]), fflush (stdout), perror (""), exit (1);
  while (!feof (f)) {
    memset (s, EOS, sizeof (s));
    fgets (s, sizeof (s) - 1, f);
    strcpy (matches, "\\2@\\1.UUCP");
    if ((match = re_strcmp (argv[1], s, matches)) > 0)
      printf ("MATCHES:%s\n", matches),
      printf ("%s", s);
    else if (match < 0)
      printf ("Error in regular expression\n"),
      exit (1);
  }
  fclose (f);
  exit (0);
}
#endif

/*
  Check 'subject' againt the 'regexpr'. Return 1 on match, 0 if no match,
  or -1 on error. To negegate a regular expression precede it with '~';
  multiple regular expressions are separated by '|' or '&' (logical OR and
  AND) and may be grouped with LGROUPCH and RGROUPCH. To escape the key
  characters LGROUPCH * ? | & RGROUPCH use \.
*/
int re_strcmp(regexpr, subject, result)
char	*regexpr, *subject, *result;
{
  char *re, s[MAXLENGTH], _re[MAXLENGTH], matches [1024];
  int i;
#ifndef egrep /* Nide */
  char *readdr; int op;
  regexp *cmp;
#else
  static regexp *cmp = NULL; /* re_strcmp is not recursive (Nide) */
#endif


#ifndef egrep
  strcpy (s, regexpr);
/*###		Do not escape ed(1) special characters.
  escape_re (s);
*/
  readdr = re = convert_re (s);
#else /* Mod Nide, regexpr can be NULL if defined(egrep) */
  re = regexpr; /* s[] not used */
#endif
#ifndef egrep
  do { /* ends with <<1>> */
    if ((op = isop (re, readdr))) {
      if (new_op (op)) {
	eval ();	/* Empty stacks */
	return -1;
      }
      ++re;
    }
    else
#endif
    {
#ifdef egrep
      if(re != NULL){ /* Nide */
#endif
	i = 0;
	while (*re != EOS && i < MAXLENGTH
#ifndef egrep
	       && !isop (re, readdr)
#endif
	      )
	  _re [i++] = *re,
	  ++re;
	if (i == MAXLENGTH)	/* Overflow */
	  return -1;
	_re [i] = EOS;
#ifdef egrep /* Nide */
	if(cmp != NULL) free(cmp);
#endif
	if (! (cmp = (regexp *) regcomp (_re))) {
#ifndef test
	  sprintf (s, "RE %s: %s\n", _re, regerr);
#ifndef noreport /* Add Nide */
	  report_progress (report, s, 1);
#endif
#else
	  printf ("RE %s: %s\n", _re, regerr);
#endif
	  free ((char *) regerr);
#ifndef egrep
	  eval ();	/* Empty stacks */
	  free ((char *) readdr);
#endif
	  return -1;
	}
#ifdef egrep /* Nide */
      }
#endif
      if (regexec (cmp, subject)) {
	if (result)
	  regsub (cmp, result, matches),
	  strcpy (result, matches);
#ifndef egrep
	free ((regexp *) cmp);
	push (1);
#else
	return 1;
#endif
      }
      else {
#ifndef egrep
	free ((regexp *) cmp);
	push (0);
#else
	return 0;
#endif
      }
    }
#ifndef egrep
  } while (*re != EOS); /* <<1>> */
  free ((char *) readdr);
  return eval ();
#endif
}

/*
  Scan 's' and escape the following characters: [ ] < > { } , ; . ^ $ + -
  for regular expression matching.

  To be used if one does not want ed(1) style pattern matching. Currently, this
  function is unused.
*/
#ifndef egrep
void	escape_re(s)
char	*s;
{
  char *r;

  while (*s != EOS) {
    switch (*s) {
    case '[': case ']': case '<': case '>': case '+': case '-':
    case ';': case ',': case '.': case '^': case '$': case '{': case '}':
      r = s + strlen (s);	/* Start from the end */
      while (r != s)
	*(r + 1) = *r,
	--r;
      *(r + 1) = *r;
      *r = '\\';
      ++s;
      break;
    }
    ++s;
  }
}

/*
  Possibly convert the wild characters * and ? to ed(1) regular expressions.
  Handle escaped characters.
*/
char	*convert_re(re)
char	*re;
{
  char *r, *b = re;
  int i = 0;

  r = (char *) malloc (sizeof (char));
  while (*re != EOS) {
    prevch (re, b);	/* See if *re is literal */
    switch (*re) {
    case '\\':
      if (re != b && literal)
	r = (char *) realloc (r, (i + 2) * sizeof (char)),
	strncpy (r + i, "\\\\", 2),
	i += 2;
      break;
    case '?':		/* Match one character */
     if (re != b && literal)	/* Literal */
	r = (char *) realloc (r, (i + 1) * sizeof (char)),
	*(r + i) = *re,
	++i;
      else
	r = (char *) realloc (r, (i + strlen (QMARK)) * sizeof (char)),
	strncpy (r + i, QMARK, strlen (QMARK)),
	i += strlen (QMARK);
      break;
    case '*':		/* Match multiple characters */
      if (re != b && literal)	/* Literal */
	r = (char *) realloc (r, (i + 2) * sizeof (char)),
	strncpy (r + i, "\\*", 2),
	i += 2;
      else
	r = (char *) realloc (r, (i + strlen (STAR)) * sizeof (char)),
	strncpy (r + i, STAR, strlen (STAR)),
	i += strlen (STAR);
      break;
    default:
      if (re != b && literal)	   /* carry over */
	r = (char *) realloc (r, (i + 2) * sizeof (char)),
	*(r + i) = *(re - 1),
	*(r + i + 1) = *re,
	i += 2;
      else
	r = (char *) realloc (r, (i + 1) * sizeof (char)),
	*(r + i) = *re,
	++i;
    }
    ++re;
  }
  r = (char *) realloc (r, (i + 1) * sizeof (char));
  *(r + i) = EOS;
  return r;
}

/*
  Return the in-coming priority of an operator.
*/
int	icp(op)
int	op;
{
  if (op == OR || op == AND) return 1;
  if (op == LPAREN || op == NOT) return 4;
  return 0;
}

/*
  Push a new operand onto OPERAND_STACK.
*/
int	push(val)
int	val;
{
  OPERAND_STACK *s;

  s = (OPERAND_STACK *) malloc (sizeof (OPERAND_STACK));
  s->val = val;
  s->prev = val_top;
  s->next = NULL;
  if (val_top)
    val_top->next = s;
  val_top = s;
  return val;
}

/*
  Pop an operand from OPERAND_STACK. Return 0 or 1, or -1 on error.
*/
int	pop()
{
  int val;

  if (!val_top)	/* Empty stack */
    return -1;
  val = val_top->val;
  if (val_top->prev)
    val_top = val_top->prev,
    free ((OPERAND_STACK *) val_top->next),
    val_top->next = NULL;
  else
    free ((OPERAND_STACK *) val_top),
    val_top = NULL;
  return val;
}

/*
  Perform a boolean operation and return the result, or -1 on error.
*/
int	do_op(op, val1, val2)
int	op, val1, val2;
{
  if (val1 < 0 || val2 < 0)
    return -1;
  if (op == OR)
    return val1 | val2;
  if (op == AND)
    return val1 & val2;
  return !val1;
}

/*
  Pop an operator. An operator will always be present.
*/
void	pop_op()
{
  OPERATOR_STACK *s;

  s = op_top;
  free ((OPERATOR_STACK *) s);
  op_top = op_top->prev;
  if (op_top)
    op_top->next = NULL;
}

/*
  Process a new operator: push it anyway, or push it after popping other
  operators with higher priority. Return -1 on error condition.
*/
int	new_op(op)
int	op;
{
  OPERATOR_STACK *s;
  int res;

  if (op < OR)
    return -1;
  if (op == RPAREN) {
    do {	/* Pop and process operators till LPAREN */
      if (!op_top)
	return -1;
      if (op_top->op != LPAREN) {
	if (op_top->op == OR || op_top->op == AND)
	  res = push (do_op (op_top->op, pop (), pop ()));
	else	/* NOT */
	  res = push (do_op (op_top->op, pop (), 0));
	if (res < 0)	/* Error with operands */
	  return res;
	pop_op ();	/* Pop processed operator */
	if (!op_top)
	  return -1;
      }
    } while (op_top->op != LPAREN);
    pop_op ();	/* LPAREN */
  }
  else {	/* Push new operator */
    while (op_top && op_top->isp >= icp (op)) {	/* Process op w/ > priority */
      if (op_top->op == OR || op_top->op == AND)
	res = push (do_op (op_top->op, pop (), pop ()));
      else    /* NOT */
	res = push (do_op (op_top->op, pop (), 0));
      if (res < 0)
	return res;
      pop_op ();
    }
    s = (OPERATOR_STACK *) malloc (sizeof (OPERATOR_STACK));
    s->op = op;
    if (op == OR) s->isp = 1;
    else if (op == AND) s->isp = 1;
    else if (op == NOT) s->isp = 2;
    else s->isp = 0;
    s->prev = op_top;
    s->next = NULL;
    if (op_top)
      op_top->next = s;
    op_top = s;
  }
  return 0;
}

/*
  Process the remaining operators and values. Return -1 on error.
*/
int	eval()
{
  OPERATOR_STACK *op;
  OPERAND_STACK *val;
  int res;

  if (!op_top && !val_top)
    return 0;
  while (op_top) {
    if (op_top->op != AND && op_top->op != OR && op_top->op != NOT) {
      if (op_top->op < OR || op_top->op > RPAREN)
#ifndef test
#ifndef noreport /* Add Nide */
	report_progress (report, "Internal error: unexpected op", 1);
#endif
#else
	printf ("Internal error: unexpected op %d\n", op_top->op);
#endif
      return -1;
    }
    if (op_top->op == OR || op_top->op == AND)
      res = push (do_op (op_top->op, pop (), pop ()));
    else	/* NOT */
      res = push (do_op (op_top->op, pop (), 0));
    if (res < 0)
      return res;
    op = op_top;
    free ((OPERATOR_STACK *) op);
    op_top = op_top->prev;
  }
  if (!val_top) {
#ifndef test
#ifndef noreport /* Add Nide */
    report_progress (report, "Internal error: empty operand stack", 1);
#endif
#else
    printf ("Internal error: empty operand stack\n");
#endif
    return -1;
  }
  return pop ();
}

/*
  Return the operator id if indeed 's' points to an operator. Return -1
  on error.
*/
int	isop(s, b)
char	*s, *b;
{
  char pch, nch;

  pch = prevch (s, b);
  if (literal) return 0;
  nch = nextch (s);
  if (*s == LGROUPCH)
    if ((pch == '|' || pch == '&' || pch == '~' || pch == LGROUPCH) &&
	!pliteral && ((nch != '|' && nch != '&' && nch != EOS) || nliteral))
      return LPAREN;
    else
      return -1;
  if (*s == RGROUPCH)
    if (((pch != '|' && pch != '&' && pch != '~') || pliteral) && !nliteral &&
	(nch == RGROUPCH || nch == '&' || nch == '|' || nch == EOS))
      return RPAREN;
    else
      return -1;
  if (*s == '~')
    if ((pch == LGROUPCH || pch == '|' || pch == '&' || pch == '~') &&
	!pliteral &&
	((nch != '|' && nch != '&' && nch != RGROUPCH && nch != EOS) ||
	 nliteral))
      return NOT;
    else
      return -1;
  if (*s == '|' || *s == '&')
    if (((pch != '~' && pch != '|' && pch != '&') || pliteral) &&
	((nch != '|' && nch != '&' && nch != RGROUPCH &&
	  nch != EOS) || nliteral))
      return (*s == '|' ? OR : AND);
    else
      return -1;
  return 0;
}

/*
  Return the previous character from the current position in the
  string. Set 'pliteral' to 1 if that previous character is escaped with \ and
  'literal' to 1 if the current character is escaped with \.
  In the comments, ^ means "beginning of the string", ? matches any character
  except \, * matches absolutely any character and x is the current position.
*/
int	prevch(s, b) /*split by Nide */
char	*s, *b;
{
  return
    ((s) <= (b) ?
     (pliteral = literal = 0, *(s)) : /* ^x */
     (((s) - 1) == (b) ?
      (*((s) - 1) == '\\' ? (pliteral = !(literal = 1), *((s) - 1)) /* ^\x */
       : (pliteral = literal = 0, *((s) - 1))) /* ^?x */
      : (*(s) == '\\' ?
	 (*((s) - 2) == '\\' ? (pliteral = !(literal = 0), *((s) - 1)) /* \*\ */
	  : (*((s) - 1) == '\\' ? (pliteral = !(literal = 1), *((s) - 2)) /* ?\\ */
	     : (pliteral = literal = 0, *((s) - 1)))) /* ??\ */
	 : (prevch2(s, b)))));
}

int	prevch2(s, b)
char	*s, *b;
{
    return (*((s) - 1) == '\\' ?
	    (*((s) - 2) == '\\' ?
	     (((s) - 2) == (b) ? (pliteral = !(literal = 0), *((s) - 1)) /* ^\\x */
	      : (*((s) - 3) == '\\' ? (pliteral = literal = 1,*((s) - 2)) /* \\\x */
		 : (pliteral = !(literal = 0), *((s) - 1)))) /* ?\\x */
	     : (((s) - 2) == (b) ? (pliteral = !(literal = 1),*((s) - 2)) /* ^?\x */
		: (*((s) - 3) == '\\' ? (pliteral = literal = 1, *((s) - 2)) /* \?\x */
		   : (pliteral = !(literal = 1),  *((s) - 2))))) /* ??\x */
	    : (prevch3(s, b)));
}

int	prevch3(s, b)
char	*s, *b;
{
       return (*((s) - 2) == '\\' ? (pliteral = !(literal = 0),*((s) - 1)) /* \?x */
	       : (pliteral = literal = 0, *((s) - 1))); /* ??x */
}

/*
  Return the next character from the current position. The current character
  is guarranteed not to be a literal. In the comments below * matches
  absolutely any character, ? matches anything but EOS, $ is the EOS and
  x marks the current position. Also set 'nliteral' to 1 if the next character
  is escaped with \.
*/
int	nextch(s)
char	*s;
{
  return
    (*((s) + 1) == EOS ? (nliteral = 0, EOS) /* x$ */
     : (*((s) + 1) == '\\' ?
	(*((s) + 2) == EOS ? (nliteral = 0, *((s) + 1)) /* x\$ */
	 : (nliteral = 1, *((s) + 2))) /* x\? */
	: (nliteral = 0, *((s) + 1)))); /* x* */
}
#endif
-regexp.c
/*
 * regcomp and regexec -- regsub and regerror are elsewhere
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 * Beware that some of this code is subtly aware of the way operator
 * precedence is structured in regular expressions.  Serious changes in
 * regular-expression syntax might require a total rethink.
 */

/*
 * The "internal use only" fields in regexp.h are present to pass info from
 * compile to execute that permits the execute phase to run lots faster on
 * simple cases.  They are:
 *
 * regstart	char that must begin a match; '\0' if none obvious
 * reganch	is the match anchored (at beginning-of-line only)?
 * regmust	string (pointer into program) that match must include, or NULL
 * regmlen	length of regmust string
 *
 * Regstart and reganch permit very fast decisions on suitable starting points
 * for a match, cutting down the work a lot.  Regmust permits fast rejection
 * of lines that cannot possibly match.  The regmust tests are costly enough
 * that regcomp() supplies a regmust only if the r.e. contains something
 * potentially expensive (at present, the only such thing detected is * or +
 * at the start of the r.e., which can involve a lot of backup).  Regmlen is
 * supplied because the test in regexec() needs it and regcomp() is computing
 * it anyway.
 */

/*
 * Structure for regexp "program".  This is essentially a linear encoding
 * of a nondeterministic finite-state machine (aka syntax charts or
 * "railroad normal form" in parsing technology).  Each node is an opcode
 * plus a "next" pointer, possibly plus an operand.  "Next" pointers of
 * all nodes except BRANCH implement concatenation; a "next" pointer with
 * a BRANCH on both ends of it is connecting two alternatives.	(Here we
 * have one of the subtle syntax dependencies:	an individual BRANCH (as
 * opposed to a collection of them) is never concatenated with anything
 * because of operator precedence.)  The operand of some types of node is
 * a literal string; for others, it is a node leading into a sub-FSM.  In
 * particular, the operand of a BRANCH node is the first node of the branch.
 * (NB this is *not* a tree structure:	the tail of the branch connects
 * to the thing following the set of BRANCHes.)  The opcodes are:
 */

/* definition	number	opnd?	meaning */
#define	END	0	/* no	End of program. */
#define	BOL	1	/* no	Match "" at beginning of line. */
#define	EOL	2	/* no	Match "" at end of line. */
#define	ANY	3	/* no	Match any one character. */
#define	ANYOF	4	/* str	Match any character in this string. */
#define	ANYBUT	5	/* str	Match any character not in this string. */
#define	BRANCH	6	/* node	Match this alternative, or the next... */
#define	BACK	7	/* no	Match "", "next" ptr points backward. */
#define	EXACTLY	8	/* str	Match this string. */
#define	NOTHING	9	/* no	Match empty string. */
#define	STAR	10	/* node	Match this (simple) thing 0 or more times. */
#define	PLUS	11	/* node	Match this (simple) thing 1 or more times. */
#define	OPEN	20	/* no	Mark this point in input as start of #n. */
			/*	OPEN+1 is number 1, etc. */
#define	CLOSE	30	/* no	Analogous to OPEN. */

/*
 * Opcode notes:
 *
 * BRANCH	The set of branches constituting a single choice are hooked
 *		together with their "next" pointers, since precedence prevents
 *		anything being concatenated to any individual branch.  The
 *		"next" pointer of the last BRANCH in a choice points to the
 *		thing following the whole choice.  This is also where the
 *		final "next" pointer of each individual branch points; each
 *		branch starts with the operand node of a BRANCH node.
 *
 * BACK		Normal "next" pointers all implicitly point forward; BACK
 *		exists to make loop structures possible.
 *
 * STAR,PLUS	'?', and complex '*' and '+', are implemented as circular
 *		BRANCH structures using BACK.  Simple cases (one character
 *		per match) are implemented with STAR and PLUS for speed
 *		and to minimize recursive plunges.
 *
 * OPEN,CLOSE	...are numbered at compile time.
 */

/*
 * A node is one char of opcode followed by two chars of "next" pointer.
 * "Next" pointers are stored as two 8-bit pieces, high order first.  The
 * value is a positive offset from the opcode of the node containing it.
 * An operand, if any, simply follows the node.  (Note that much of the
 * code generation knows about this implicit relationship.)
 *
 * Using two bytes for the "next" pointer is vast overkill for most things,
 * but allows patterns to get big without disasters.
 */
#define	OP(p)	(*(p))
#define	NEXT(p)	(((*((p)+1)&0377)<<8) + *((p)+2)&0377)
#define	OPERAND(p)	((p) + 3)

/*
 * See regmagic.h for one further detail of program structure.
 */


/*
 * Utility definitions.
 */
#ifndef CHARBITS
#define	UCHARAT(p)	((int)*(unsigned char *)(p))
#else
#define	UCHARAT(p)	((int)*(p)&CHARBITS)
#endif

#define	FAIL(m)	{ regerror(m); return(NULL); }
#define	ISMULT(c)	((c) == '*' || (c) == '+' || (c) == '?')
#define	META	"^$.[()|?+*\\"

/*
 * Flags to be passed up and down.
 */
#define	HASWIDTH	01	/* Known never to match null string. */
#define	SIMPLE		02	/* Simple enough to be STAR/PLUS operand. */
#define	SPSTART		04	/* Starts with * or +. */
#define	WORST		0	/* Worst case. */

/*
 * Global work variables for regcomp().
 */
static char *regparse;		/* Input-scan pointer. */
static int regnpar;		/* () count. */
static char regdummy;
static char *regcode;		/* Code-emit pointer; &regdummy = don't. */
static long regsize;		/* Code size. */

/*
 * Forward declarations for regcomp()'s friends.
 */
#ifndef STATIC
#define	STATIC	static
#endif
STATIC	char	*reg ARGDECL((int, int *));
STATIC	char	*regbranch ARGDECL((int *));
STATIC	char	*regpiece ARGDECL((int *));
STATIC	char	*regatom ARGDECL((int *));
STATIC	char	*regnode ARGDECL((char));
STATIC	char	*regnext ARGDECL((char *));
STATIC	void	regc ARGDECL((char));
STATIC	void	reginsert ARGDECL((char, char *));
STATIC	void	regtail ARGDECL((char *, char *));
STATIC	void	regoptail ARGDECL((char *, char *));
#ifdef STRCSPN
STATIC	int	strcspn();
#endif

/*
 - regcomp - compile a regular expression into internal code
 *
 * We can't allocate space until we know how big the compiled form will be,
 * but we can't compile it (and thus know how big it is) until we've got a
 * place to put the code.  So we cheat:  we compile it twice, once with code
 * generation turned off and size counting turned on, and once "for real".
 * This also means that we don't allocate space until we are sure that the
 * thing really will compile successfully, and we never have to move the
 * code and thus invalidate pointers into it.  (Note that it has to be in
 * one piece because free() must be able to free it all.)
 *
 * Beware that the optimization-preparation code in here knows about some
 * of the structure of the compiled regexp.
 */
regexp *
regcomp(exp)
char *exp;
{
	register regexp *r;
	register char *scan;
	register char *longest;
	register int len;
	int flags;
	/* extern char *malloc(); */

	if (exp == NULL)
		FAIL("NULL argument");

	/* First pass: determine size, legality. */
	regparse = exp;
	regnpar = 1;
	regsize = 0L;
	regcode = &regdummy;
	regc(MAGIC);
	if (reg(0, &flags) == NULL)
		return(NULL);

	/* Small enough for pointer-storage convention? */
	if (regsize >= 32767L)		/* Probably could be 65535L. */
		FAIL("regexp too big");

	/* Allocate space. */
	r = (regexp *)malloc(sizeof(regexp) + (unsigned)regsize);
	if (r == NULL)
		FAIL("out of space");

	/* Second pass: emit code. */
	regparse = exp;
	regnpar = 1;
	regcode = r->program;
	regc(MAGIC);
	if (reg(0, &flags) == NULL)
		return(NULL);

	/* Dig out information for optimizations. */
	r->regstart = '\0';	/* Worst-case defaults. */
	r->reganch = 0;
	r->regmust = NULL;
	r->regmlen = 0;
	scan = r->program+1;			/* First BRANCH. */
	if (OP(regnext(scan)) == END) {		/* Only one top-level choice. */
		scan = OPERAND(scan);

		/* Starting-point info. */
		if (OP(scan) == EXACTLY)
			r->regstart = *OPERAND(scan);
		else if (OP(scan) == BOL)
			r->reganch++;

		/*
		 * If there's something expensive in the r.e., find the
		 * longest literal string that must appear and make it the
		 * regmust.  Resolve ties in favor of later strings, since
		 * the regstart check works with the beginning of the r.e.
		 * and avoiding duplication strengthens checking.  Not a
		 * strong reason, but sufficient in the absence of others.
		 */
		if (flags&SPSTART) {
			longest = NULL;
			len = 0;
			for (; scan != NULL; scan = regnext(scan))
				if (OP(scan) == EXACTLY && strlen(OPERAND(scan)) >= len) {
					longest = OPERAND(scan);
					len = strlen(OPERAND(scan));
				}
			r->regmust = longest;
			r->regmlen = len;
		}
	}

	return(r);
}

/*
 - reg - regular expression, i.e. main body or parenthesized thing
 *
 * Caller must absorb opening parenthesis.
 *
 * Combining parenthesis handling with the base level of regular expression
 * is a trifle forced, but the need to tie the tails of the branches to what
 * follows makes it hard to avoid.
 */
static char *
reg(paren, flagp)
int paren;			/* Parenthesized? */
int *flagp;
{
	register char *ret;
	register char *br;
	register char *ender;
	register int parno;
	int flags;

	*flagp = HASWIDTH;	/* Tentatively. */

	/* Make an OPEN node, if parenthesized. */
	if (paren) {
		if (regnpar >= NSUBEXP)
			FAIL("too many ()");
		parno = regnpar;
		regnpar++;
		ret = regnode((char)(OPEN+parno));
	} else
		ret = NULL;

	/* Pick up the branches, linking them together. */
	br = regbranch(&flags);
	if (br == NULL)
		return(NULL);
	if (ret != NULL)
		regtail(ret, br);	/* OPEN -> first. */
	else
		ret = br;
	if (!(flags&HASWIDTH))
		*flagp &= ~HASWIDTH;
	*flagp |= flags&SPSTART;
	while (*regparse == '|') {
		regparse++;
		br = regbranch(&flags);
		if (br == NULL)
			return(NULL);
		regtail(ret, br);	/* BRANCH -> BRANCH. */
		if (!(flags&HASWIDTH))
			*flagp &= ~HASWIDTH;
		*flagp |= flags&SPSTART;
	}

	/* Make a closing node, and hook it on the end. */
	ender = regnode((paren) ? (char)(CLOSE+parno) : END);
	regtail(ret, ender);

	/* Hook the tails of the branches to the closing node. */
	for (br = ret; br != NULL; br = regnext(br))
		regoptail(br, ender);

	/* Check for proper termination. */
	if (paren && *regparse++ != ')') {
		FAIL("unmatched ()");
	} else if (!paren && *regparse != '\0') {
		if (*regparse == ')') {
			FAIL("unmatched ()");
		} else
			FAIL("junk on end");	/* "Can't happen". */
		/* NOTREACHED */
	}

	return(ret);
}

/*
 - regbranch - one alternative of an | operator
 *
 * Implements the concatenation operator.
 */
static char *
regbranch(flagp)
int *flagp;
{
	register char *ret;
	register char *chain;
	register char *latest;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	ret = regnode(BRANCH);
	chain = NULL;
	while (*regparse != '\0' && *regparse != '|' && *regparse != ')') {
		latest = regpiece(&flags);
		if (latest == NULL)
			return(NULL);
		*flagp |= flags&HASWIDTH;
		if (chain == NULL)	/* First piece. */
			*flagp |= flags&SPSTART;
		else
			regtail(chain, latest);
		chain = latest;
	}
	if (chain == NULL)	/* Loop ran zero times. */
		(void) regnode(NOTHING);

	return(ret);
}

/*
 - regpiece - something followed by possible [*+?]
 *
 * Note that the branching code sequences used for ? and the general cases
 * of * and + are somewhat optimized:  they use the same NOTHING node as
 * both the endmarker for their branch list and the body of the last branch.
 * It might seem that this node could be dispensed with entirely, but the
 * endmarker role is not redundant.
 */
static char *
regpiece(flagp)
int *flagp;
{
	register char *ret;
	register char op;
	register char *next;
	int flags;

	ret = regatom(&flags);
	if (ret == NULL)
		return(NULL);

	op = *regparse;
	if (!ISMULT(op)) {
		*flagp = flags;
		return(ret);
	}

	if (!(flags&HASWIDTH) && op != '?')
		FAIL("*+ operand could be empty");
	*flagp = (op != '+') ? (WORST|SPSTART) : (WORST|HASWIDTH);

	if (op == '*' && (flags&SIMPLE))
		reginsert(STAR, ret);
	else if (op == '*') {
		/* Emit x* as (x&|), where & means "self". */
		reginsert(BRANCH, ret);			/* Either x */
		regoptail(ret, regnode(BACK));		/* and loop */
		regoptail(ret, ret);			/* back */
		regtail(ret, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '+' && (flags&SIMPLE))
		reginsert(PLUS, ret);
	else if (op == '+') {
		/* Emit x+ as x(&|), where & means "self". */
		next = regnode(BRANCH);			/* Either */
		regtail(ret, next);
		regtail(regnode(BACK), ret);		/* loop back */
		regtail(next, regnode(BRANCH));		/* or */
		regtail(ret, regnode(NOTHING));		/* null. */
	} else if (op == '?') {
		/* Emit x? as (x|) */
		reginsert(BRANCH, ret);			/* Either x */
		regtail(ret, regnode(BRANCH));		/* or */
		next = regnode(NOTHING);		/* null. */
		regtail(ret, next);
		regoptail(ret, next);
	}
	regparse++;
	if (ISMULT(*regparse))
		FAIL("nested *?+");

	return(ret);
}

/*
 - regatom - the lowest level
 *
 * Optimization:  gobbles an entire sequence of ordinary characters so that
 * it can turn them into a single node, which is smaller to store and
 * faster to run.  Backslashed characters are exceptions, each becoming a
 * separate node; the code is simpler that way and it's not worth fixing.
 */
static char *
regatom(flagp)
int *flagp;
{
	register char *ret;
	int flags;

	*flagp = WORST;		/* Tentatively. */

	switch (*regparse++) {
	case '^':
		ret = regnode(BOL);
		break;
	case '$':
		ret = regnode(EOL);
		break;
	case '.':
		ret = regnode(ANY);
		*flagp |= HASWIDTH|SIMPLE;
		break;
	case '[': {
			register int class;
			register int classend;

			if (*regparse == '^') {	/* Complement of range. */
				ret = regnode(ANYBUT);
				regparse++;
			} else
				ret = regnode(ANYOF);
			if (*regparse == ']' || *regparse == '-')
				regc(*regparse++);
			while (*regparse != '\0' && *regparse != ']') {
				if (*regparse == '-') {
					regparse++;
					if (*regparse == ']' || *regparse == '\0')
						regc('-');
					else {
						class = UCHARAT(regparse-2)+1;
						classend = UCHARAT(regparse);
						if (class > classend+1)
							FAIL("invalid [] range");
						for (; class <= classend; class++)
							regc((char)class);
						regparse++;
					}
				} else
					regc(*regparse++);
			}
			regc('\0');
			if (*regparse != ']')
				FAIL("unmatched []");
			regparse++;
			*flagp |= HASWIDTH|SIMPLE;
		}
		break;
	case '(':
		ret = reg(1, &flags);
		if (ret == NULL)
			return(NULL);
		*flagp |= flags&(HASWIDTH|SPSTART);
		break;
	case '\0':
	case '|':
	case ')':
		FAIL("internal urp");	/* Supposed to be caught earlier. */
		break;
	case '?':
	case '+':
	case '*':
		FAIL("?+* follows nothing");
		break;
	case '\\':
		if (*regparse == '\0')
			FAIL("trailing \\");
		ret = regnode(EXACTLY);
		regc(*regparse++);
		regc('\0');
		*flagp |= HASWIDTH|SIMPLE;
		break;
	default: {
			register int len;
			register char ender;

			regparse--;
			len = strcspn(regparse, META);
			if (len <= 0)
				FAIL("internal disaster");
			ender = *(regparse+len);
			if (len > 1 && ISMULT(ender))
				len--;		/* Back off clear of ?+* operand. */
			*flagp |= HASWIDTH;
			if (len == 1)
				*flagp |= SIMPLE;
			ret = regnode(EXACTLY);
			while (len > 0) {
				regc(*regparse++);
				len--;
			}
			regc('\0');
		}
		break;
	}

	return(ret);
}

/*
 - regnode - emit a node
 */
static char *			/* Location. */
regnode(op)
char op;
{
	register char *ret;
	register char *ptr;

	ret = regcode;
	if (ret == &regdummy) {
		regsize += 3;
		return(ret);
	}

	ptr = ret;
	*ptr++ = op;
	*ptr++ = '\0';		/* Null "next" pointer. */
	*ptr++ = '\0';
	regcode = ptr;

	return(ret);
}

/*
 - regc - emit (if appropriate) a byte of code
 */
static void
regc(b)
char b;
{
	if (regcode != &regdummy)
		*regcode++ = b;
	else
		regsize++;
}

/*
 - reginsert - insert an operator in front of already-emitted operand
 *
 * Means relocating the operand.
 */
static void
reginsert(op, opnd)
char op;
char *opnd;
{
	register char *src;
	register char *dst;
	register char *place;

	if (regcode == &regdummy) {
		regsize += 3;
		return;
	}

	src = regcode;
	regcode += 3;
	dst = regcode;
	while (src > opnd)
		*--dst = *--src;

	place = opnd;		/* Op node, where operand used to be. */
	*place++ = op;
	*place++ = '\0';
	*place++ = '\0';
}

/*
 - regtail - set the next-pointer at the end of a node chain
 */
static void
regtail(p, val)
char *p;
char *val;
{
	register char *scan;
	register char *temp;
	register int offset;

	if (p == &regdummy)
		return;

	/* Find last node. */
	scan = p;
	for (;;) {
		temp = regnext(scan);
		if (temp == NULL)
			break;
		scan = temp;
	}

	if (OP(scan) == BACK)
		offset = scan - val;
	else
		offset = val - scan;
	*(scan+1) = (char)((offset>>8)&0377);
	*(scan+2) = (char)(offset&0377);
}

/*
 - regoptail - regtail on operand of first argument; nop if operandless
 */
static void
regoptail(p, val)
char *p;
char *val;
{
	/* "Operandless" and "op != BRANCH" are synonymous in practice. */
	if (p == NULL || p == &regdummy || OP(p) != BRANCH)
		return;
	regtail(OPERAND(p), val);
}

/*
 * regexec and friends
 */

/*
 * Global work variables for regexec().
 */
static char *reginput;		/* String-input pointer. */
static char *regbol;		/* Beginning of input, for ^ check. */
static char **regstartp;	/* Pointer to startp array. */
static char **regendp;		/* Ditto for endp. */

/*
 * Forwards.
 */
STATIC	int	regtry ARGDECL((regexp *, char *));
STATIC	int	regmatch ARGDECL((char *));
STATIC	int	regrepeat ARGDECL((char *));

#ifdef DEBUG
int regnarrate = 0;
void	regdump();
STATIC	char	*regprop();
#endif

/*
 - regexec - match a regexp against a string
 */
int
regexec(prog, string)
register regexp *prog;
register char *string;
{
	register char *s;
	extern char *strchr();

	/* Be paranoid... */
	if (prog == NULL || string == NULL) {
		regerror("NULL parameter");
		return(0);
	}

	/* Check validity of program. */
	if (UCHARAT(prog->program) != MAGIC) {
		regerror("corrupted program");
		return(0);
	}

	/* If there is a "must appear" string, look for it. */
	if (prog->regmust != NULL) {
		s = string;
		while ((s = strchr(s, prog->regmust[0])) != NULL) {
			if (strncmp(s, prog->regmust, prog->regmlen) == 0)
				break;	/* Found it. */
			s++;
		}
		if (s == NULL)	/* Not present. */
			return(0);
	}

	/* Mark beginning of line for ^ . */
	regbol = string;

	/* Simplest case:  anchored match need be tried only once. */
	if (prog->reganch)
		return(regtry(prog, string));

	/* Messy cases:  unanchored match. */
	s = string;
	if (prog->regstart != '\0')
		/* We know what char it must start with. */
		while ((s = strchr(s, prog->regstart)) != NULL) {
			if (regtry(prog, s))
				return(1);
			s++;
		}
	else
		/* We don't -- general case. */
		do {
			if (regtry(prog, s))
				return(1);
		} while (*s++ != '\0');

	/* Failure. */
	return(0);
}

/*
 - regtry - try match at specific point
 */
static int			/* 0 failure, 1 success */
regtry(prog, string)
regexp *prog;
char *string;
{
	register int i;
	register char **sp;
	register char **ep;

	reginput = string;
	regstartp = prog->startp;
	regendp = prog->endp;

	sp = prog->startp;
	ep = prog->endp;
	for (i = NSUBEXP; i > 0; i--) {
		*sp++ = NULL;
		*ep++ = NULL;
	}
	if (regmatch(prog->program + 1)) {
		prog->startp[0] = string;
		prog->endp[0] = reginput;
		return(1);
	} else
		return(0);
}

/*
 - regmatch - main matching routine
 *
 * Conceptually the strategy is simple:  check to see whether the current
 * node matches, call self recursively to see whether the rest matches,
 * and then act accordingly.  In practice we make some effort to avoid
 * recursion, in particular by going through "ordinary" nodes (that don't
 * need to know whether the rest of the match failed) by a loop instead of
 * by recursion.
 */
static int			/* 0 failure, 1 success */
regmatch(prog)
char *prog;
{
	register char *scan;	/* Current node. */
	char *next;		/* Next node. */
	extern char *strchr();

	scan = prog;
#ifdef DEBUG
	if (scan != NULL && regnarrate)
		fprintf(stderr, "%s(\n", regprop(scan));
#endif
	while (scan != NULL) {
#ifdef DEBUG
		if (regnarrate)
			fprintf(stderr, "%s...\n", regprop(scan));
#endif
		next = regnext(scan);

		switch (OP(scan)) {
		case BOL:
			if (reginput != regbol)
				return(0);
			break;
		case EOL:
			if (*reginput != '\0')
				return(0);
			break;
		case ANY:
			if (*reginput == '\0')
				return(0);
			reginput++;
			break;
		case EXACTLY: {
				register int len;
				register char *opnd;

				opnd = OPERAND(scan);
				/* Inline the first character, for speed. */
				if (*opnd != *reginput)
					return(0);
				len = strlen(opnd);
				if (len > 1 && strncmp(opnd, reginput, len) != 0)
					return(0);
				reginput += len;
			}
			break;
		case ANYOF:
			if (*reginput == '\0' || strchr(OPERAND(scan), *reginput) == NULL)
				return(0);
			reginput++;
			break;
		case ANYBUT:
			if (*reginput == '\0' || strchr(OPERAND(scan), *reginput) != NULL)
				return(0);
			reginput++;
			break;
		case NOTHING:
			break;
		case BACK:
			break;
		case OPEN+1:
		case OPEN+2:
		case OPEN+3:
		case OPEN+4:
		case OPEN+5:
		case OPEN+6:
		case OPEN+7:
		case OPEN+8:
		case OPEN+9: {
				register int no;
				register char *save;

				no = OP(scan) - OPEN;
				save = reginput;

				if (regmatch(next)) {
					/*
					 * Don't set startp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regstartp[no] == NULL)
						regstartp[no] = save;
					return(1);
				} else
					return(0);
			}
			break;
		case CLOSE+1:
		case CLOSE+2:
		case CLOSE+3:
		case CLOSE+4:
		case CLOSE+5:
		case CLOSE+6:
		case CLOSE+7:
		case CLOSE+8:
		case CLOSE+9: {
				register int no;
				register char *save;

				no = OP(scan) - CLOSE;
				save = reginput;

				if (regmatch(next)) {
					/*
					 * Don't set endp if some later
					 * invocation of the same parentheses
					 * already has.
					 */
					if (regendp[no] == NULL)
						regendp[no] = save;
					return(1);
				} else
					return(0);
			}
			break;
		case BRANCH: {
				register char *save;

				if (OP(next) != BRANCH)		/* No choice. */
					next = OPERAND(scan);	/* Avoid recursion. */
				else {
					do {
						save = reginput;
						if (regmatch(OPERAND(scan)))
							return(1);
						reginput = save;
						scan = regnext(scan);
					} while (scan != NULL && OP(scan) == BRANCH);
					return(0);
					/* NOTREACHED */
				}
			}
			break;
		case STAR:
		case PLUS: {
				register char nextch;
				register int no;
				register char *save;
				register int minn;
		 /* min -> minn, because LSI-C has min(). Nide */

				/*
				 * Lookahead to avoid useless match attempts
				 * when we know what character comes next.
				 */
				nextch = '\0';
				if (OP(next) == EXACTLY)
					nextch = *OPERAND(next);
				minn = (OP(scan) == STAR) ? 0 : 1;
				save = reginput;
				no = regrepeat(OPERAND(scan));
				while (no >= minn) {
					/* If it could work, try it. */
					if (nextch == '\0' || *reginput == nextch)
						if (regmatch(next))
							return(1);
					/* Couldn't or didn't -- back up. */
					no--;
					reginput = save + no;
				}
				return(0);
			}
			break;
		case END:
			return(1);	/* Success! */
			break;
		default:
			regerror("memory corruption");
			return(0);
			break;
		}

		scan = next;
	}

	/*
	 * We get here only if there's trouble -- normally "case END" is
	 * the terminating point.
	 */
	regerror("corrupted pointers");
	return(0);
}

/*
 - regrepeat - repeatedly match something simple, report how many
 */
static int
regrepeat(p)
char *p;
{
	register int count = 0;
	register char *scan;
	register char *opnd;

	scan = reginput;
	opnd = OPERAND(p);
	switch (OP(p)) {
	case ANY:
		count = strlen(scan);
		scan += count;
		break;
	case EXACTLY:
		while (*opnd == *scan) {
			count++;
			scan++;
		}
		break;
	case ANYOF:
		while (*scan != '\0' && strchr(opnd, *scan) != NULL) {
			count++;
			scan++;
		}
		break;
	case ANYBUT:
		while (*scan != '\0' && strchr(opnd, *scan) == NULL) {
			count++;
			scan++;
		}
		break;
	default:		/* Oh dear.  Called inappropriately. */
		regerror("internal foulup");
		count = 0;	/* Best compromise. */
		break;
	}
	reginput = scan;

	return(count);
}

/*
 - regnext - dig the "next" pointer out of a node
 */
static char *
regnext(p)
register char *p;
{
	register int offset;

	if (p == &regdummy)
		return(NULL);

	offset = NEXT(p);
	if (offset == 0)
		return(NULL);

	if (OP(p) == BACK)
		return(p-offset);
	else
		return(p+offset);
}

#ifdef DEBUG

STATIC	char	*regprop();

/*
 - regdump - dump a regexp onto stdout in vaguely comprehensible form
 */
void
regdump(r)
regexp *r;
{
	register char *s;
	register char op = EXACTLY;	/* Arbitrary non-END op. */
	register char *next;
	extern char *strchr();


	s = r->program + 1;
	while (op != END) {	/* While that wasn't END last time... */
		op = OP(s);
		printf("%2d%s", s-r->program, regprop(s));	/* Where, what. */
		next = regnext(s);
		if (next == NULL)		/* Next ptr. */
			printf("(0)");
		else
			printf("(%d)", (s-r->program)+(next-s));
		s += 3;
		if (op == ANYOF || op == ANYBUT || op == EXACTLY) {
			/* Literal string, where present. */
			while (*s != '\0') {
				putchar(*s);
				s++;
			}
			s++;
		}
		putchar('\n');
	}

	/* Header fields of interest. */
	if (r->regstart != '\0')
		printf("start `%c' ", r->regstart);
	if (r->reganch)
		printf("anchored ");
	if (r->regmust != NULL)
		printf("must have \"%s\"", r->regmust);
	printf("\n");
}

/*
 - regprop - printable representation of opcode
 */
static char *
regprop(op)
char *op;
{
	register char *p;
	static char buf[50];

	(void) strcpy(buf, ":");

	switch (OP(op)) {
	case BOL:
		p = "BOL";
		break;
	case EOL:
		p = "EOL";
		break;
	case ANY:
		p = "ANY";
		break;
	case ANYOF:
		p = "ANYOF";
		break;
	case ANYBUT:
		p = "ANYBUT";
		break;
	case BRANCH:
		p = "BRANCH";
		break;
	case EXACTLY:
		p = "EXACTLY";
		break;
	case NOTHING:
		p = "NOTHING";
		break;
	case BACK:
		p = "BACK";
		break;
	case END:
		p = "END";
		break;
	case OPEN+1:
	case OPEN+2:
	case OPEN+3:
	case OPEN+4:
	case OPEN+5:
	case OPEN+6:
	case OPEN+7:
	case OPEN+8:
	case OPEN+9:
		sprintf(buf+strlen(buf), "OPEN%d", OP(op)-OPEN);
		p = NULL;
		break;
	case CLOSE+1:
	case CLOSE+2:
	case CLOSE+3:
	case CLOSE+4:
	case CLOSE+5:
	case CLOSE+6:
	case CLOSE+7:
	case CLOSE+8:
	case CLOSE+9:
		sprintf(buf+strlen(buf), "CLOSE%d", OP(op)-CLOSE);
		p = NULL;
		break;
	case STAR:
		p = "STAR";
		break;
	case PLUS:
		p = "PLUS";
		break;
	default:
		regerror("corrupted opcode");
		break;
	}
	if (p != NULL)
		(void) strcat(buf, p);
	return(buf);
}
#endif

/*
 * The following is provided for those people who do not have strcspn() in
 * their C libraries.  They should get off their butts and do something
 * about it; at least one public-domain implementation of those (highly
 * useful) string routines has been published on Usenet.
 */
#ifdef STRCSPN
/*
 * strcspn - find length of initial segment of s1 consisting entirely
 * of characters not from s2
 */

static int
strcspn(s1, s2)
char *s1;
char *s2;
{
	register char *scan1;
	register char *scan2;
	register int count;

	count = 0;
	for (scan1 = s1; *scan1 != '\0'; scan1++) {
		for (scan2 = s2; *scan2 != '\0';)	/* ++ moved down. */
			if (*scan1 == *scan2++)
				return(count);
		count++;
	}
	return(count);
}
#endif
-regsub.c
/*
 * regsub
 *
 *	Copyright (c) 1986 by University of Toronto.
 *	Written by Henry Spencer.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 */

#ifndef CHARBITS
#define	UCHARAT(p)	((int)*(unsigned char *)(p))
#else
#define	UCHARAT(p)	((int)*(p)&CHARBITS)
#endif

/*
 - regsub - perform substitutions after a regexp match
 */
void
regsub(prog, source, dest)
regexp *prog;
char *source;
char *dest;
{
	register char *src;
	register char *dst;
	register char c;
	register int no;
	register int len;
	extern char *strncpy();

	if (prog == NULL || source == NULL || dest == NULL) {
		regerror("NULL parm to regsub");
		return;
	}
	if (UCHARAT(prog->program) != MAGIC) {
		regerror("damaged regexp fed to regsub");
		return;
	}

	src = source;
	dst = dest;
	while ((c = *src++) != '\0') {
		if (c == '&')
			no = 0;
		else if (c == '\\' && '0' <= *src && *src <= '9')
			no = *src++ - '0';
		else
			no = -1;

		if (no < 0) {	/* Ordinary character. */
			if (c == '\\' && (*src == '\\' || *src == '&'))
				c = *src++;
			*dst++ = c;
		}
		else if (prog->startp[no] != NULL && prog->endp[no] != NULL) {
			len = prog->endp[no] - prog->startp[no];
			(void) strncpy(dst, prog->startp[no], len);
			dst += len;
			if (len != 0 && *(dst-1) == '\0') {		/* strncpy hit NUL. */
				regerror("damaged match string");
				return;
			}
		}
	}
	*dst++ = '\0';
}
-regerror.c
/*
  Tasos Kotsikonas (8/20/92): Changed function so that it stores the error
  message to a global variable.
*/

char *regerr;

void
regerror(s)
char *s;
{
  regerr = (char *) malloc ((strlen (s) + 1) * sizeof (char));
  strcpy (regerr, s);
}

/* ==================== original makefile
# This is a modified Makefile from Henry Spencer's original.
# Things you might want to put in ENV and LENV:
# -Dvoid=int		compilers that don't do void
# -DCHARBITS=0377	compilers that don't do unsigned char
# -DSTATIC=extern	compilers that don't like "static foo();" as forward decl
# -DSTRCSPN		library does not have strcspn()
# -Dstrchr=index	library does not have strchr()
# -DERRAVAIL		have utzoo-compatible error() function and friends
ENV=
LENV=

# Things you might want to put in TEST:
# -DDEBUG		debugging hooks
# -I.			regexp.h from current directory, not /usr/include
# -Dtest		to run independetly
TEST=-I.

# Things you might want to put in PROF:
# -Dstatic=""		make everything global so profiler can see it.
# -p			profiler
PROF=

CC=cl
CFLAGS=/c /F1800 $(ENV) $(TEST) $(PROF)
LINTFLAGS=$(LENV) $(TEST) -ha
LDFLAGS=

OBJ=regexp.obj regsub.obj regex.obj regerror.obj
LSRC=regexp.c regsub.c regerror.c

all:	regex.lib

regex.lib:	$(OBJ)
	lib regex-+regexp-+regsub-+regerror-+regex, nul, regex
	command /C type nul > regex.bak
	command /C del regex.bak

regexp.obj:	regexp.c regexp.h regmagic.h
	$(CC) $(CFLAGS) regexp.c regexp.h regmagic.h
regsub.obj:	regsub.c regexp.h regmagic.h
	$(CC) $(CFLAGS) regsub.c regexp.h regmagic.h
regerror.obj:	regerror.c
	$(CC) $(CFLAGS) regerror.c
regex.obj:	regex.c regexp.h
	$(CC) $(CFLAGS) regex.c regexp.h

clean:
	rm -f *.obj regex.lib
 ==================== */
